# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import base64
import hashlib
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import cbor2
import ccf.receipt
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.x509 import load_der_x509_certificate
from pycose.messages import Sign1Message
from pycose.messages.cosebase import CoseBase

from . import crypto

if TYPE_CHECKING:
    from .verify import ServiceParameters

HEADER_PARAM_TREE_ALGORITHM = "tree_alg"
TREE_ALGORITHM_CCF = "CCF"


def hdr_as_dict(phdr: dict) -> dict:
    """
    Return a representation of a list of COSE header parameters that
    is amenable to pretty-printing.
    """

    def display(item):
        if hasattr(item, "__name__"):
            return item.__name__
        if type(item) is bytes:
            return item.hex()
        return item

    # Decode KID into a 'readable' text string if present.
    hdr_dict = {display(k): display(v) for k, v in phdr.items()}
    if hdr_dict.get("KID"):
        hdr_dict["KID"] = bytes.fromhex(hdr_dict["KID"]).decode()

    return hdr_dict


@dataclass
class LeafInfo:
    internal_hash: bytes
    internal_data: bytes

    @classmethod
    def from_cose_obj(cls, cose_obj: list) -> "LeafInfo":
        return cls(*cose_obj)

    def digest(self, claims_digest: bytes) -> bytes:
        internal_data_digest = hashlib.sha256(self.internal_data).digest()
        return hashlib.sha256(
            self.internal_hash + internal_data_digest + claims_digest
        ).digest()


class ReceiptContents(ABC):
    @abstractmethod
    def verify(self, tbs: bytes, service: "ServiceParameters"):
        pass

    @abstractmethod
    def as_dict(self) -> dict:
        pass

    @classmethod
    def from_cose_obj(self, headers: dict, cose_obj: Any) -> "ReceiptContents":
        if headers.get(HEADER_PARAM_TREE_ALGORITHM) == TREE_ALGORITHM_CCF:
            return CCFReceiptContents(
                cose_obj[0],
                cose_obj[1],
                cose_obj[2],
                LeafInfo.from_cose_obj(cose_obj[3]),
            )
        else:
            raise ValueError("unsupported tree algorithm, cannot decode receipt")


@dataclass
class CCFReceiptContents(ReceiptContents):
    signature: bytes
    node_certificate: bytes
    inclusion_proof: list
    leaf_info: LeafInfo

    def root(self, claims_digest: bytes) -> bytes:
        leaf = self.leaf_info.digest(claims_digest).hex()

        proof = []
        for [left, hash_] in self.inclusion_proof:
            if left:
                proof.append({"left": hash_.hex()})
            else:
                proof.append({"right": hash_.hex()})

        return bytes.fromhex(ccf.receipt.root(leaf, proof))

    def verify(self, tbs: bytes, service: "ServiceParameters"):
        if service.tree_algorithm != "CCF":
            raise ValueError("treeAlgorithm must be CCF")
        if service.signature_algorithm != "ES256":
            raise ValueError("signatureAlgorithm must be ES256")

        claims_digest = hashlib.sha256(tbs).digest()

        service_cert = load_der_x509_certificate(service.certificate)
        node_cert = load_der_x509_certificate(self.node_certificate)
        if not isinstance(node_cert.public_key(), ec.EllipticCurvePublicKey):
            raise ValueError("Invalid node public key algorithm")

        root = self.root(claims_digest).hex()

        # The CCF module expects a base64 signature, in ASN1/DER format.
        signature = crypto.convert_p1363_signature_to_dss(self.signature)
        b64signature = base64.b64encode(signature).decode()

        ccf.receipt.verify(root, b64signature, node_cert)
        ccf.receipt.check_endorsement(node_cert, service_cert)

    def as_dict(self) -> dict:
        """
        Return a representation of the receipt that is amenable to
        pretty-printing.
        """
        proof = []
        for [left, h] in self.inclusion_proof:
            if left:
                proof.append({"left": h.hex()})
            else:
                proof.append({"right": h.hex()})

        return {
            "signature": base64.b64encode(self.signature).decode("ascii"),
            "node_certificate": base64.b64encode(self.node_certificate).decode("ascii"),
            "inclusion_proof": proof,
            "leaf_info": {
                "internal_hash": self.leaf_info.internal_hash.hex(),
                "internal_data": self.leaf_info.internal_data.decode("ascii"),
            },
        }


@dataclass
class Receipt:
    phdr_encoded: bytes
    phdr: dict
    contents: ReceiptContents

    @classmethod
    def from_cose_obj(cls, cose_obj: list) -> "Receipt":
        phdr_encoded = cose_obj.pop(0)
        phdr = CoseBase._parse_header(cbor2.loads(phdr_encoded), True)
        contents = ReceiptContents.from_cose_obj(phdr, cose_obj.pop(0))

        return Receipt(phdr_encoded, phdr, contents)

    @classmethod
    def decode(cls, data: bytes) -> "Receipt":
        return cls.from_cose_obj(cbor2.loads(data))  # type: ignore[arg-type]

    def countersign_structure(self, claim: Sign1Message) -> bytes:
        context = "CounterSignatureV2"
        countersign_structure = [
            context,
            claim.phdr_encoded,
            self.phdr_encoded,
            b"",  # no external AAD
            claim.payload,
            [claim.signature],
        ]
        return cbor2.dumps(countersign_structure)

    def verify(self, claim: Sign1Message, service_params: "ServiceParameters"):
        tbs = self.countersign_structure(claim)
        self.contents.verify(tbs, service_params)

    def as_dict(self) -> dict:
        """
        Return a representation of the protected headers that is amenable
        to pretty-printing.
        """
        return {
            "protected": hdr_as_dict(self.phdr),
            "contents": self.contents.as_dict(),
        }



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import argparse
import json
import os
import shutil
import subprocess
import tempfile
from pathlib import Path

DID_WEB_PREFIX = "did:web:"
GITHUB_IO_SUFFIX = ".github.io"
DID_FILENAME = "did.json"
WELLKNOWN_DIR = ".well-known"


def run(*args):
    print(" ".join(args))
    subprocess.run(args)


def upload_did_web_doc_to_github(did_path: Path):
    did_path = did_path.resolve()
    with open(did_path) as f:
        doc = json.load(f)
    did = doc["id"]
    assert did.startswith(DID_WEB_PREFIX), f"DID must start with {DID_WEB_PREFIX}"
    domain = did.replace(DID_WEB_PREFIX, "")
    assert domain.endswith(
        GITHUB_IO_SUFFIX
    ), f"did:web domain must end with {GITHUB_IO_SUFFIX}"
    gh_user = domain.replace(GITHUB_IO_SUFFIX, "")
    repo_url = f"https://github.com/{gh_user}/{domain}"

    with tempfile.TemporaryDirectory() as tmpdir:
        repo_path = os.path.join(tmpdir, domain)
        run("git", "clone", repo_url, repo_path)
        os.chdir(repo_path)
        os.makedirs(WELLKNOWN_DIR, exist_ok=True)
        did_repo_path = os.path.join(WELLKNOWN_DIR, DID_FILENAME)
        shutil.copyfile(did_path, did_repo_path)
        # Adding .nojekyll allows a fresh GitHub pages site to serve files.
        Path(".nojekyll").touch()
        run("git", "add", ".nojekyll", did_repo_path)
        run("git", "commit", "-m", f"update {did_repo_path}")
        run("git", "push")


def cli(fn):
    parser = fn(
        description="Upload a DID document for did:web:user.github.io to "
        "an existing user/user.github.io GitHub repository"
    )
    parser.add_argument("path", type=Path, help="Path to did.json file")

    def cmd(args):
        upload_did_web_doc_to_github(args.path)

    parser.set_defaults(func=cmd)

    return parser


if __name__ == "__main__":
    parser = cli(argparse.ArgumentParser)
    args = parser.parse_args()
    args.func(args)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import hashlib
from dataclasses import dataclass
from http import HTTPStatus
from typing import TYPE_CHECKING, List, Union

import cbor2
from pycose.messages import CoseMessage, Sign1Message
from pycose.messages.cosebase import CoseBase

if TYPE_CHECKING:
    from .client import BaseClient

from .crypto import COSE_HEADER_PARAM_FEED, COSE_HEADER_PARAM_ISSUER
from .receipt import ReceiptContents, hdr_as_dict
from .verify import ServiceParameters


class bitvector:
    def __init__(self, data):
        self.data = data

    def __getitem__(self, i):
        assert i < self.size()
        return bool(self.data[i // 8] & (0x80 >> (i % 8)))

    def prefix(self, n):
        assert n < self.size()

        data = [0] * len(self.data)
        for i in range(n // 8):
            data[i] = self.data[i]

        padding = 0x80 >> (n % 8)
        mask = ~(padding - 1)

        i = n // 8
        data[i] = (self.data[i] & mask) | padding

        return bytes(data)

    def size(self):
        return len(self.data) * 8

    def bit_count(self):
        return sum(self[i] for i in range(self.size()))


@dataclass
class Path:
    positions: bytes
    hashes: List[bytes]

    def __post_init__(self):
        positions = bitvector(self.positions)
        if positions.bit_count() != len(self.hashes):
            raise ValueError(
                f"Expected {positions.bit_count()} hashes, got {len(self.hashes)} instead"
            )

    def hash(self, index: bytes, leaf: bytes) -> bytes:
        positions = bitvector(self.positions)
        hashes = reversed(self.hashes)
        index_bits = bitvector(index)

        current = leaf
        for i in reversed(range(256)):
            if positions[i]:
                node = hashlib.sha256(index_bits.prefix(i))
                if index_bits[i]:
                    node.update(next(hashes))
                    node.update(current)
                else:
                    node.update(current)
                    node.update(next(hashes))
                current = node.digest()
        return current


@dataclass
class ReadReceipt:
    tree_headers: dict
    leaf_headers: dict
    tree_hdr_encoded: bytes
    leaf_hdr_encoded: bytes
    proof: Path
    tree_receipt: ReceiptContents

    @classmethod
    def from_cose_obj(cls, cose_obj: list) -> "ReadReceipt":
        tree_hdr_encoded = cose_obj.pop(0)
        leaf_hdr_encoded = cose_obj.pop(0)
        path = Path(*cose_obj.pop(0))

        tree_hdr = CoseBase._parse_header(cbor2.loads(tree_hdr_encoded), True)
        leaf_hdr = CoseBase._parse_header(cbor2.loads(leaf_hdr_encoded), True)
        tree_receipt = ReceiptContents.from_cose_obj(tree_hdr, cose_obj.pop(0))

        return ReadReceipt(
            tree_hdr, leaf_hdr, tree_hdr_encoded, leaf_hdr_encoded, path, tree_receipt
        )

    @classmethod
    def decode(cls, data: bytes):
        return cls.from_cose_obj(cbor2.loads(data))  # type: ignore[arg-type]

    def verify(
        self, claim: Union[Sign1Message, bytes], service_params: ServiceParameters
    ):
        if isinstance(claim, bytes):
            claim_ = CoseMessage.decode(claim)
            assert isinstance(claim_, Sign1Message)
            claim = claim_

        tbs = self.tree_tbs(claim)
        self.tree_receipt.verify(tbs, service_params)

    def tree_tbs(self, claim: Sign1Message) -> bytes:
        context = "SCITT"
        structure = [context, self.tree_hdr_encoded, self.root(claim)]
        return cbor2.dumps(structure)

    def root(self, claim: Sign1Message) -> bytes:
        index = self.claim_index(claim)
        leaf = self.leaf_hash(index, claim)
        return self.proof.hash(index, leaf)

    @classmethod
    def claim_index(cls, claim: Sign1Message):
        issuer = claim.phdr[COSE_HEADER_PARAM_ISSUER]
        feed = claim.phdr.get(COSE_HEADER_PARAM_FEED, "")
        return hashlib.sha256(cbor2.dumps([issuer, feed])).digest()

    def leaf_hash(self, index: bytes, claim: Sign1Message) -> bytes:
        digest = hashlib.sha256(self.leaf_tbs(claim)).digest()
        return hashlib.sha256(index + digest).digest()

    def leaf_tbs(self, claim: Sign1Message) -> bytes:
        context = "CounterSignatureV2"
        structure = [
            context,
            claim.phdr_encoded,
            self.leaf_hdr_encoded,
            b"",  # no external AAD
            claim.payload,
            [claim.signature],
        ]
        return cbor2.dumps(structure)

    def as_dict(self) -> dict:
        return {
            "tree_headers": hdr_as_dict(self.tree_headers),
            "leaf_headers": hdr_as_dict(self.leaf_headers),
            "proof": {
                "positions": self.proof.positions.hex(),
                "hashes": [h.hex() for h in self.proof.hashes],
            },
            "tree_receipt": self.tree_receipt.as_dict(),
        }


@dataclass
class TreeReceipt:
    headers_encoded: bytes
    headers: dict
    root: bytes
    receipt_contents: ReceiptContents

    @classmethod
    def decode(cls, data: bytes) -> "TreeReceipt":
        cose_obj = cbor2.loads(data)
        headers_encoded = cose_obj.pop(0)  # type: ignore[attr-defined]
        headers = CoseBase._parse_header(cbor2.loads(headers_encoded), True)
        root = cose_obj.pop(0)  # type: ignore[attr-defined]
        receipt_contents = ReceiptContents.from_cose_obj(headers, cose_obj.pop(0))  # type: ignore[attr-defined]
        return TreeReceipt(headers_encoded, headers, root, receipt_contents)

    @property
    def upper_bound_seqno(self) -> int:
        return self.headers["upper_bound_seqno"]


class PrefixTreeClient:
    client: "BaseClient"

    def __init__(self, client: "BaseClient"):
        self.client = client

    def flush(self):
        def is_indexed(data: bytes, upper_bound_seqno: int) -> bool:
            return TreeReceipt.decode(data).upper_bound_seqno >= upper_bound_seqno

        info = self.client.post("/prefix_tree/flush", wait_for_confirmation=True).json()

        # The confirmation we get from the flush is not quite enough for the
        # results to be visible, since it may take some time for the indexer to
        # process the commit.
        #
        # Query the indexer until its upper bound is greater or equal to what
        # we got when flushing. We also need to handle the NoPrefixTree 404 error,
        # which can happen on a fresh service.
        self.client.get_historical(
            "/prefix_tree",
            retry_on=[
                (HTTPStatus.NOT_FOUND, "NoPrefixTree"),
                lambda r: r.is_success
                and not is_indexed(r.content, info["upper_bound"]),
            ],
        )
        return info

    def debug(self) -> dict:
        return self.client.get("/prefix_tree/debug").json()

    def get_read_receipt(self, issuer: str, feed: str, *, decode=True):
        response = self.client.get_historical(f"/read_receipt/{issuer}/{feed}")
        if decode:
            return ReadReceipt.decode(response.content)
        else:
            return response.content



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import base64
import hashlib
from typing import Dict, Optional

from azure.identity import DefaultAzureCredential
from azure.keyvault.certificates import CertificateClient, KeyVaultCertificate
from azure.keyvault.keys import KeyClient
from azure.keyvault.keys.crypto import CryptographyClient, SignatureAlgorithm
from ccf.cose import create_cose_sign1_finish, create_cose_sign1_prepare
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from cryptography.x509 import load_pem_x509_certificate

from pyscitt.client import MemberAuthenticationMethod

from . import crypto

ALGORITHMS = {
    256: ("ES256", "sha256"),
    384: ("ES384", "sha384"),
    521: ("ES512", "sha384"),
}


class KeyVaultSignClient(MemberAuthenticationMethod):
    """MemberIdentity implementation that uses Azure Key Vault."""

    def __init__(self, akv_sign_configuration_dict: dict):
        self._vault_name = akv_sign_configuration_dict["keyVaultName"]
        self._vault_url = f"https://{self._vault_name}.vault.azure.net"
        self._identity_certificate_name = akv_sign_configuration_dict["certificateName"]
        self._identity_certificate_version = akv_sign_configuration_dict[
            "certificateVersion"
        ]
        self.credential = DefaultAzureCredential()
        cert_client = CertificateClient(
            vault_url=self._vault_url, credential=self.credential
        )
        self.cert = self._encode_certificate(
            cert_client.get_certificate_version(
                certificate_name=self._identity_certificate_name,
                version=self._identity_certificate_version,
            )
        )

    @staticmethod
    def _encode_certificate(cert: KeyVaultCertificate) -> str:
        assert isinstance(cert.cer, bytearray)
        decoded = base64.b64encode(cert.cer).decode()
        cert_pem = f"-----BEGIN CERTIFICATE-----\n{decoded}\n-----END CERTIFICATE-----"
        return cert_pem

    def _get_crypto_client(self):
        key_client = KeyClient(vault_url=self._vault_url, credential=self.credential)
        key = key_client.get_key(
            name=self._identity_certificate_name,
            version=self._identity_certificate_version,
        )
        crypto_client = CryptographyClient(key, credential=self.credential)
        return crypto_client

    def cose_sign(self, data: bytes, cose_headers: Optional[Dict] = None) -> bytes:
        """Generates a COSE payload for the specified request with the specified headers.

        https://microsoft.github.io/CCF/main/use_apps/issue_commands.html#signing

        :param data: The intended body for the HTTP request.
        :type data: bytes
        :param cose_headers: The headers to include in the COSE payload.
        :type cose_headers: Optional[Dict]

        :return: The full payload, with signature, to be sent to CCF.
        :rtype: bytes
        """

        assert self.cert, "No identity public certificate available for this identity"

        # Prepare the COSE payload
        tbs = create_cose_sign1_prepare(
            payload=data,
            cert_pem=self.cert,
            additional_protected_header=cose_headers,
        )

        # Sign the unsigned COSE payload using Key Vault
        digest_to_sign = base64.b64decode(tbs["value"].encode())
        algorithm = SignatureAlgorithm(tbs["alg"])
        crypto_client = self._get_crypto_client()
        sign_result = crypto_client.sign(
            algorithm=SignatureAlgorithm(algorithm), digest=digest_to_sign
        )

        # Finish to prepare the COSE payload
        return create_cose_sign1_finish(
            payload=data,
            cert_pem=self.cert,
            signature=base64.urlsafe_b64encode(sign_result.signature).decode(),
            additional_protected_header=cose_headers,
        )

    def http_sign(self, data: bytes):
        """
        Generates a HTTP signing payload for the specified request with the specified headers.

        https://microsoft.github.io/CCF/main/use_apps/issue_commands.html#signing

        :param data: The intended body for the HTTP request.
        :type data: bytes

        :return: The full payload, with signature, to be sent to CCF.
        :rtype: bytes
        """

        assert self.cert, "No identity public certificate available for this identity"

        crypto_client = self._get_crypto_client()
        cert = load_pem_x509_certificate(self.cert.encode("ascii"))
        pub_key = cert.public_key()
        assert isinstance(pub_key, (EllipticCurvePublicKey))
        key_size = pub_key.curve.key_size
        signature_algorithm, hash_algorithm = ALGORITHMS[key_size]

        digest_to_sign = hashlib.new(hash_algorithm, data).digest()
        sign_result = crypto_client.sign(
            algorithm=SignatureAlgorithm(signature_algorithm), digest=digest_to_sign
        )
        return crypto.convert_p1363_signature_to_dss(sign_result.signature)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import argparse
from pathlib import Path
from typing import Optional

from ..did import Resolver
from ..verify import DIDResolverTrustStore, StaticTrustStore, TrustStore, verify_receipt


def validate_cose_with_receipt(
    cose_path: Path,
    receipt_path: Optional[Path],
    service_trust_store_path: Optional[Path],
    verify_did_connection: bool = False,
):
    cose = cose_path.read_bytes()

    if receipt_path is None:
        receipt = None
    else:
        receipt = receipt_path.read_bytes()

    service_trust_store: TrustStore
    if not service_trust_store_path:
        service_trust_store = DIDResolverTrustStore(
            resolver=Resolver(verify=verify_did_connection)
        )
    else:
        service_trust_store = StaticTrustStore.load(service_trust_store_path)

    verify_receipt(cose, service_trust_store, receipt)
    print(f"COSE document is valid: {cose_path}")


def cli(fn):
    parser = fn(description="Validate a COSE_Sign1 document using a receipt")
    parser.add_argument("cose", type=Path, help="Path to COSE file")
    parser.add_argument(
        "--receipt",
        type=Path,
        help="Optional path to receipt, otherwise read from COSE headers",
    )
    parser.add_argument(
        "--service-trust-store",
        type=Path,
        help="""Optional folder containing JSON parameter files of SCITT services to trust,
        otherwise use DID resolver and expect DID in receipt""",
    )
    parser.add_argument(
        "--verify-did-connection",
        action="store_true",
        default=False,
        help="For DID resolution, whether to verify the connection to the DID server. Default is False.",
    )

    def cmd(args):
        validate_cose_with_receipt(
            args.cose,
            args.receipt,
            args.service_trust_store,
            args.verify_did_connection,
        )

    parser.set_defaults(func=cmd)

    return parser


if __name__ == "__main__":
    parser = cli(argparse.ArgumentParser)
    args = parser.parse_args()
    args.func(args)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import argparse
from pathlib import Path
from typing import Optional

from ..client import Client
from ..verify import StaticTrustStore, verify_receipt
from .client_arguments import add_client_arguments, create_client


def submit_signed_claimset(
    client: Client,
    path: Path,
    receipt_path: Optional[Path],
    service_trust_store_path: Optional[Path],
    skip_confirmation: bool,
):
    if path.suffix != ".cose":
        raise ValueError("unsupported file extension")

    with open(path, "rb") as f:
        signed_claimset = f.read()

    if skip_confirmation:
        pending = client.submit_claim(signed_claimset, skip_confirmation=True)
        print(f"Submitted {path} as operation {pending.operation_tx}")
        print("Confirmation of submission was skipped! Claim may not be registered.")
        return

    submission = client.submit_claim(signed_claimset)
    print(f"Submitted {path} as transaction {submission.tx}")

    if receipt_path:
        with open(receipt_path, "wb") as f:
            f.write(submission.raw_receipt)
        print(f"Received {receipt_path}")

    if service_trust_store_path:
        service_trust_store = StaticTrustStore.load(service_trust_store_path)
        verify_receipt(
            signed_claimset,
            receipt=submission.receipt,
            service_trust_store=service_trust_store,
        )


def cli(fn):
    parser = fn(
        description="Submit signed claimset to a SCITT CCF Ledger and retrieve receipt"
    )
    add_client_arguments(parser, with_auth_token=True)
    parser.add_argument("path", type=Path, help="Path to signed claimset file")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--receipt", type=Path, help="Output path to receipt file")
    group.add_argument(
        "--skip-confirmation",
        action="store_true",
        help="Don't wait for confirmation or a receipt",
    )
    parser.add_argument(
        "--service-trust-store",
        type=Path,
        help="Folder containing JSON parameter files of SCITT services to trust, used to verify the claim",
    )

    def cmd(args):
        client = create_client(args)
        submit_signed_claimset(
            client,
            args.path,
            args.receipt,
            args.service_trust_store,
            args.skip_confirmation,
        )

    parser.set_defaults(func=cmd)

    return parser


if __name__ == "__main__":
    parser = cli(argparse.ArgumentParser)
    args = parser.parse_args()
    args.func(args)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import base64
import hashlib
import json
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from http import HTTPStatus
from typing import Any, Dict, Iterable, Literal, Optional, TypeVar, Union, overload
from urllib.parse import urlencode

import httpx
from loguru import logger as LOG

from . import crypto
from .governance import GovernanceClient
from .prefix_tree import PrefixTreeClient
from .receipt import Receipt
from .verify import ServiceParameters

CCF_TX_ID_HEADER = "x-ms-ccf-transaction-id"


class SigningType(Enum):
    """Types of signatures supported by CCF.

    https://microsoft.github.io/CCF/main/governance/hsm_keys.html#signing-governance-requests
    """

    COSE = "COSE"
    HTTP = "HTTP"


class MemberAuthenticationMethod(ABC):
    cert: str

    @abstractmethod
    def http_sign(self, data: bytes) -> bytes:
        """
        Generates a HTTP signing payload for the specified request with the specified headers.

        https://microsoft.github.io/CCF/main/use_apps/issue_commands.html#signing

        :param data: The intended body for the HTTP request.
        :type data: bytes

        :return: The full payload, with signature, to be sent to CCF.
        :rtype: bytes
        """

    @abstractmethod
    def cose_sign(self, data: bytes, cose_headers: Optional[Dict] = None) -> bytes:
        """Generates a COSE payload for the specified request with the specified headers.

        https://microsoft.github.io/CCF/main/use_apps/issue_commands.html#signing

        :param data: The intended body for the HTTP request.
        :type data: bytes
        :param cose_headers: The headers to include in the COSE payload.
        :type cose_headers: Optional[Dict]

        :return: The full payload, with signature, to be sent to CCF.
        :rtype: bytes
        """


# copied from CCF/tests/infra/clients.py
class HttpSig(httpx.Auth):
    member_auth_client: MemberAuthenticationMethod
    requires_request_body = True

    def __init__(self, member_auth_client: MemberAuthenticationMethod):
        self.member_auth_client = member_auth_client
        self.key_id = crypto.get_cert_fingerprint(member_auth_client.cert)

    def auth_flow(self, request):
        body_digest = base64.b64encode(hashlib.sha256(request.content).digest()).decode(
            "ascii"
        )
        request.headers["digest"] = f"SHA-256={body_digest}"
        string_to_sign = "\n".join(
            [
                f"(request-target): {request.method.lower()} {request.url.raw_path.decode('utf-8')}",
                f"digest: SHA-256={body_digest}",
                f"content-length: {len(request.content)}",
            ]
        ).encode("utf-8")

        signature = self.member_auth_client.http_sign(string_to_sign)
        b64signature = base64.b64encode(signature).decode("ascii")
        request.headers["authorization"] = (
            f'Signature keyId="{self.key_id}",algorithm="hs2019",headers="(request-target) digest content-length",signature="{b64signature}"'
        )
        yield request


@dataclass
class ServiceError(Exception):
    headers: httpx.Headers
    code: str
    message: str

    def __str__(self):
        return f"{self.code}: {self.message}"


SelfClient = TypeVar("SelfClient", bound="BaseClient")


# When running functional tests, we may send identical proposals that,
# if signed within the same second, they may hit the ProposalReplay
# protection error from CCF.
# We use a custom clock that can be manually advanced, same as
# implemented by CCF, to avoid sleeping for 1 second between
# identical proposals.
# Source: https://github.com/microsoft/CCF/blob/d6efe6664045968dc4f191b9ae672e686f05279b/tests/infra/clients.py#L40
class OffSettableSecondsSinceEpoch:
    offset = 0

    def count(self):
        return self.offset + int(datetime.now().timestamp())

    def advance(self, amount=1):
        LOG.info(f"Advancing clock by {amount} seconds")
        self.offset += amount


CLOCK = OffSettableSecondsSinceEpoch()


def cose_protected_headers(request_path: str, method: str):
    """
    Generate the COSE protected headers for CCF governance given a request path and HTTP method.

    :param request_path: The path of the request.
    :type request_path: str
    :param method: The method of the request.
    :type method: str

    :return: The COSE protected headers.
    :rtype: dict
    """

    cose_headers: Dict[str, Any] = {}

    # Set the created_at header to the current time
    cose_headers = {
        "ccf.gov.msg.created_at": CLOCK.count(),
    }

    # Set headers based on the request path and method
    if request_path.endswith("gov/ack/update_state_digest"):
        cose_headers["ccf.gov.msg.type"] = "state_digest"
    elif request_path.endswith("gov/ack"):
        cose_headers["ccf.gov.msg.type"] = "ack"
    elif request_path.endswith("gov/proposals"):
        cose_headers["ccf.gov.msg.type"] = "proposal"
    elif request_path.endswith("/ballots"):
        pid = request_path.split("/")[-2]
        cose_headers["ccf.gov.msg.type"] = "ballot"
        cose_headers["ccf.gov.msg.proposal_id"] = pid
    elif request_path.endswith("/withdraw"):
        pid = request_path.split("/")[-2]
        cose_headers["ccf.gov.msg.type"] = "withdrawal"
        cose_headers["ccf.gov.msg.proposal_id"] = pid
    elif request_path.endswith("gov/recovery_share"):
        if method == "GET":
            cose_headers["ccf.gov.msg.type"] = "encrypted_recovery_share"
        if method == "POST":
            cose_headers["ccf.gov.msg.type"] = "recovery_share"

    return cose_headers


def get_content_data(body: Optional[Union[dict, str, bytes]]) -> bytes:
    """
    Get the request body from the body parameter

    :param body: The body of the request. Can be a string, bytes, or dict.
    :type body: Optional[Union[dict, str, bytes]]

    :return: The request body
    :rtype: bytes
    """

    if isinstance(body, str):
        request_body = body.encode()
    elif isinstance(body, dict):
        request_body = json.dumps(body).encode()
    elif isinstance(body, bytes):
        request_body = body
    else:
        raise ValueError(f"Invalid body type: {type(body)}")

    return request_body


class BaseClient:
    """
    Wrapper around an HTTP client, with facilities to interact with a CCF-based
    service.

    Provides support for authentication, request signing, logging, configurable
    retry conditions, etc...
    """

    url: str
    auth_token: Optional[str]
    member_auth: Optional[MemberAuthenticationMethod]
    member_signing_type: SigningType
    wait_time: Optional[float]
    development: bool
    cacert: Optional[str]

    session: httpx.Client
    member_http_sig: Optional[HttpSig]

    def __init__(
        self,
        url: str,
        *,
        auth_token: Optional[str] = None,
        member_auth: Optional[MemberAuthenticationMethod] = None,
        member_signing_type: SigningType = SigningType.COSE,
        wait_time: Optional[float] = None,
        development: bool = False,
        cacert: Optional[str] = None,
    ):
        """
        Create a new BaseClient instance.

        auth_token:
            A bearer token for all requests made by this instance.

        member_auth:
            MemberAuthenticationMethod include A pair of certificate and private key in PEM format or AKV login identity, used to sign requests.
            Each request that needs signing must also be given the `sign_request=True` parameter.

        wait_time:
            The time to wait between retries. If None, the default wait time is used.

        member_signing_type:
            The type of signing to use for member authentication. Currently, only COSE and HTTP signing are supported.

        development:
            If true, the TLS certificate of the server will not be verified.

        cacert:
            If set and development is False, will be used in TLS verification instead of the default bundle.
        """

        # Even though these are passed to the httpx.Client and not used
        # directly, we save them so we can create a new Client with modified
        # settings in the `replace` method.
        self.url = url
        self.auth_token = auth_token
        self.member_auth = member_auth
        self.member_signing_type = member_signing_type
        self.wait_time = wait_time
        self.development = development
        self.cacert = cacert

        headers = {}
        if auth_token:
            headers["Authorization"] = "Bearer " + auth_token

        # We only create a custom HTTPX authentication instance for HTTP signing
        # because COSE signing cannot be handled that way and requires re-writing
        # the response payload.
        if member_auth and member_signing_type == SigningType.HTTP:
            self.member_http_sig = HttpSig(member_auth)
        else:
            self.member_http_sig = None

        tls_verification: Union[str, bool] = (
            cacert if cacert is not None else not development
        )

        self.session = httpx.Client(
            base_url=url, headers=headers, verify=tls_verification
        )

    def replace(self: SelfClient, **kwargs) -> SelfClient:
        """
        Create a new instance with certain parameters modified. Any parameters
        that weren't specified will be inherited from the current instance.

        The accepted keyword arguments are the same as those of the constructor.
        """
        values: dict = {
            "url": self.url,
            "auth_token": self.auth_token,
            "member_auth": self.member_auth,
            "member_signing_type": self.member_signing_type,
            "wait_time": self.wait_time,
            "development": self.development,
            "cacert": self.cacert,
        }
        values.update(kwargs)
        return self.__class__(**values)

    def request(
        self,
        method,
        url,
        *,
        retry_on=[],
        sign_request=False,
        wait_for_confirmation=False,
        **kwargs,
    ) -> httpx.Response:
        """
        Issue a request to the server.

        retry_on: a list of predicates on the response.
            Each predicate can be one of an HTTP status code, a pair of a status code and a error
            code or a callable that accepts a response object and returns True if the request
            should be retried. If any predicate matches, the request is retried after a short delay.

        sign_request:
            Sign the request with the member private key. A key and certificate must have been
            provided to the constructor.

        wait_for_confirmation:
            Wait for the transaction created by this request to be globally committed. If True, the
            client will wait for confirmation and raise an error if the transaction was rolled-back.
            Otherwise, any write to the key-value store caused by this request could be silently
            rolled-back.

        Other keyword-arguments are passed to httpx.
        """
        if sign_request:
            # Check that the request is not already signed
            if "auth" in kwargs:
                raise ValueError("Cannot use `auth` with `sign_request`")

            # Sign with COSE
            if self.member_signing_type == SigningType.COSE and self.member_auth:
                # Advance the clock to avoid ProposalReplay protection errors
                if self.development:
                    CLOCK.advance()

                # Get the COSE headers
                cose_headers = cose_protected_headers(url, method)

                def _get_data():
                    """Get the data to sign from the request keywords"""

                    # The data is either in `content` or `json` kwarg.
                    # We assume that other keywords used to pass the
                    # content for HTTPX requests are not used
                    # (e.g., such as the deprecated `data`).
                    content = kwargs.get("content")
                    if content:
                        content = get_content_data(content)

                    json_data = kwargs.get("json")
                    if json_data:
                        json_data = get_content_data(json_data)

                    # If both are specified, raise an error
                    if content and json_data:
                        raise ValueError("Cannot use both `content` and `json`")

                    # Return the data to sign
                    # If both are not specified, return an empty bytes string
                    # (e.g., for GET requests)
                    return content or json_data or b""

                # Sign the data
                payload = self.member_auth.cose_sign(_get_data(), cose_headers)

                # Set the request data and the content-type header
                kwargs["content"] = payload
                kwargs.setdefault("headers", {})["content-type"] = "application/cose"

            # Sign with HTTP signing
            elif self.member_signing_type == SigningType.HTTP and self.member_http_sig:
                kwargs["auth"] = self.member_http_sig

                if method == "GET":
                    # Content-length is necessary for signing, even on GET requests.
                    kwargs.setdefault("headers", {}).setdefault("Content-Length", "0")

            else:
                raise ValueError(f"Cannot sign request with {self.member_signing_type}")

        default_wait_time = 2
        timeout = 30
        deadline = time.monotonic() + timeout
        attempt = 1
        while True:
            response = self.session.request(method, url, **kwargs)

            log_parts = [method, url]
            if attempt > 1:
                log_parts.append(f"(attempt #{attempt})")
            log_parts.append(response.status_code)
            if not response.is_success:
                log_parts.append(response.json().get("error", {}).get("code"))
            LOG.debug(" ".join(str(p) for p in log_parts))

            for code in retry_on:
                if isinstance(code, tuple):
                    if (
                        response.status_code == code[0]
                        and response.json().get("error", {}).get("code") == code[1]
                    ):
                        break
                elif callable(code):
                    if code(response):
                        break
                else:
                    if response.status_code == code:
                        break
            else:
                break

            if self.wait_time is not None:
                wait = self.wait_time
            else:
                wait = int(response.headers.get("retry-after", default_wait_time))
            if time.monotonic() + wait > deadline:
                raise ValueError("Too many retries")

            time.sleep(wait)
            attempt += 1

        if not response.is_success:
            error = response.json()["error"]
            LOG.error(f"Request failed: {error}")
            raise ServiceError(response.headers, error["code"], error["message"])

        if wait_for_confirmation:
            self.wait_for_confirmation(response.headers[CCF_TX_ID_HEADER])

        return response

    def wait_for_confirmation(self, tx: str):
        """
        Wait until a transaction has either been globally committed or was rolled back.

        In the latter case, an exception is raised.
        """
        response = self.get(
            "/node/tx",
            params={"transaction_id": tx},
            retry_on=[lambda r: r.is_success and r.json()["status"] == "Pending"],
        )

        status = response.json()["status"]
        if status != "Committed":
            raise RuntimeError(
                f"Transaction {tx} was not committed. Status is {status}"
            )

    def get(self, *args, **kwargs) -> httpx.Response:
        return self.request("GET", *args, **kwargs)

    def post(self, *args, **kwargs) -> httpx.Response:
        return self.request("POST", *args, **kwargs)

    def get_historical(self, *args, retry_on=[], **kwargs):
        """
        Issue a request, retrying on codes commonly used by CCF applications to indicate that a
        historical query to the KV is in progress and needs to be retried.
        """
        return self.get(
            *args,
            **kwargs,
            retry_on=[
                (HTTPStatus.SERVICE_UNAVAILABLE, "TransactionNotCached"),
            ]
            + retry_on,
        )


@dataclass
class PendingSubmission:
    """
    The result of submitting a claim to the service.
    """

    operation_tx: str


@dataclass
class Submission(PendingSubmission):
    tx: str
    raw_receipt: bytes

    @property
    def seqno(self) -> int:
        view, seqno = self.tx.split(".")
        return int(seqno)

    @property
    def receipt(self) -> Receipt:
        return Receipt.decode(self.raw_receipt)


class Client(BaseClient):
    """
    Specialization of the BaseClient, aimed at interacting with a SCITT CCF ledger instance.
    """

    def get_parameters(self) -> ServiceParameters:
        return ServiceParameters.from_dict(self.get("/parameters").json())

    def get_constitution(self) -> str:
        # The endpoint returns the value as a JSON-encoded string, ie. wrapped
        # in double quotes and with all special characters escaped.
        return self.get("/gov/kv/constitution").json()

    def get_version(self) -> dict:
        return self.get("/version").json()

    def get_did_document(self, did: str) -> dict:
        # Note: This endpoint only returns data for did:web DIDs.
        return self.get(f"/did/{did}").json()["did_document"]

    @overload
    def submit_claim(
        self, claim: bytes, *, skip_confirmation: Literal[False] = False
    ) -> Submission: ...

    @overload
    def submit_claim(
        self, claim: bytes, *, skip_confirmation: Literal[True]
    ) -> PendingSubmission: ...

    def submit_claim(
        self, claim: bytes, *, skip_confirmation=False
    ) -> Union[Submission, PendingSubmission]:
        headers = {"Content-Type": "application/cose"}
        response = self.post(
            "/entries",
            headers=headers,
            content=claim,
        ).json()

        operation_id = response["operationId"]
        if skip_confirmation:
            return PendingSubmission(operation_id)
        else:
            tx = self.wait_for_operation(operation_id)
            receipt = self.get_receipt(tx, decode=False)
            return Submission(operation_id, tx, receipt)

    def wait_for_operation(self, operation: str) -> str:
        response = self.get(
            f"/operations/{operation}",
            retry_on=[lambda r: r.is_success and r.json()["status"] == "running"],
        )
        payload = response.json()

        if payload["status"] == "succeeded":
            return payload["entryId"]
        elif payload["status"] == "failed":
            error = payload["error"]
            raise ServiceError(response.headers, error["code"], error["message"])
        else:
            raise ValueError("Invalid status {}".format(payload["status"]))

    def get_operations(self):
        return self.get("/operations").json()["operations"]

    def get_claim(self, tx: str, *, embed_receipt=False) -> bytes:
        response = self.get_historical(
            f"/entries/{tx}", params={"embedReceipt": embed_receipt}
        )
        return response.content

    @overload
    def get_receipt(
        self,
        tx: str,
        *,
        operation: bool = False,
        decode: Literal[True] = True,
    ) -> Receipt: ...

    @overload
    def get_receipt(
        self, tx: str, *, operation: bool = False, decode: Literal[False]
    ) -> bytes: ...

    def get_receipt(
        self, tx: str, *, operation: bool = False, decode: bool = True
    ) -> Union[bytes, Receipt]:
        """
        Get a receipt from the ledger.

        If `operation` is true, the tx is treated as an operation ID and is
        first waited on in order to obtain the actual entry ID.
        """
        if operation:
            tx = self.wait_for_operation(tx)

        response = self.get_historical(f"/entries/{tx}/receipt")
        if decode:
            return Receipt.decode(response.content)
        else:
            return response.content

    def enumerate_claims(
        self, *, start: Optional[int] = None, end: Optional[int] = None
    ) -> Iterable[str]:
        """
        Enumerate all claims on the ledger, with an optional start and end range.

        Yields a sequence of transaction numbers. The contents and/or receipt for a given claim can
        be fetched using the `get_claim` and `get_receipt` methods.
        """
        params = {}
        if start is not None:
            params["from"] = start
        if end is not None:
            params["to"] = end

        link = f"/entries/txIds?{urlencode(params)}"

        while link:
            response = self.get(
                link,
                params=params,
                retry_on=[
                    (HTTPStatus.SERVICE_UNAVAILABLE, "IndexingInProgressRetryLater")
                ],
            )
            body = response.json()
            for tx in body["transactionIds"]:
                yield tx

            link = body.get("nextLink")

    def wait_for_network_open(self):
        self.get(
            "/node/network",
            retry_on=[lambda r: r.is_success and r.json()["service_status"] != "Open"],
        )

    def get_service_certificate(self) -> crypto.Pem:
        return self.get("/node/network").json()["service_certificate"]

    def get_previous_service_identity(self) -> crypto.Pem:
        return self.get("/node/service/previous_identity").json()[
            "previous_service_identity"
        ]

    @property
    def governance(self):
        return GovernanceClient(self)

    @property
    def prefix_tree(self):
        return PrefixTreeClient(self)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import argparse
import json
from pathlib import Path

from ..receipt import Receipt


def prettyprint_receipt(receipt_path: Path):
    with open(receipt_path, "rb") as f:
        receipt = f.read()
    parsed = Receipt.decode(receipt)
    print(json.dumps(parsed.as_dict(), indent=2))


def cli(fn):
    parser = fn(description="Pretty-print a SCITT receipt")
    parser.add_argument("receipt", type=Path, help="Path to SCITT receipt file")

    def cmd(args):
        prettyprint_receipt(args.receipt)

    parser.set_defaults(func=cmd)

    return parser


if __name__ == "__main__":
    parser = cli(argparse.ArgumentParser)
    args = parser.parse_args()
    args.func(args)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import argparse
import base64
import json
import re
from enum import Enum
from pathlib import Path
from typing import List, Optional

import pycose.headers
from azure.identity import DefaultAzureCredential
from azure.keyvault.secrets import SecretClient
from cryptography.hazmat.primitives.serialization import pkcs12
from cryptography.x509 import load_pem_x509_certificate

from pyscitt.key_vault_sign_client import KeyVaultSignClient

from .. import crypto, did


class RegistrationInfoType(Enum):
    TEXT = "text"
    BYTES = "bytes"
    INT = "int"


class RegistrationInfoArgument:
    type: RegistrationInfoType
    name: str
    content: str

    # This won't support names that contain an '=' or ':'.
    # This is probably fine for the time being, but we should have a scheme to escape those.
    PATTERN = re.compile("((?P<type>[^=:]+):)?" "(?P<name>[^=:]+)=" "(?P<content>.*)")

    def __init__(self, value: str):
        match = self.PATTERN.fullmatch(value)
        if not match:
            raise argparse.ArgumentTypeError(
                f"'{value}' is not a valid registration info argument"
            )

        type_ = match.group("type")
        if type_ is None:
            self.type = RegistrationInfoType.TEXT
        else:
            try:
                self.type = RegistrationInfoType(type_)
            except Exception as e:
                raise argparse.ArgumentTypeError(
                    f"'{type}' is not a valid registration info type"
                ) from None

        self.name = match.group("name")
        self.content = match.group("content")

    def value(self) -> crypto.RegistrationInfoValue:
        if self.content.startswith("@"):
            data = Path(self.content[1:]).read_bytes()
        else:
            data = self.content.encode("ascii")

        if self.type is RegistrationInfoType.INT:
            return int(data.decode("utf-8"))
        elif self.type is RegistrationInfoType.TEXT:
            return data.decode("utf-8")
        elif self.type is RegistrationInfoType.BYTES:
            return data


def _parse_x5c_file(x5c_path: str) -> List[str]:
    # Load x509 root certificates
    with open(x5c_path, "rb") as f:
        cacert_data = f.read()

    certs = []
    while cacert_data:
        pemcert, _, cacert_data = cacert_data.partition(b"-----END CERTIFICATE-----\n")
        pemcert += b"-----END CERTIFICATE-----\n"
        certs.append(pemcert.decode())
    return certs


def create_signer_from_arguments(
    key_path: Path,
    did_doc_path: Optional[Path],
    kid: Optional[str],
    issuer: Optional[str],
    algorithm: Optional[str],
    x5c_path: Optional[str],
) -> crypto.Signer:
    key = crypto.load_private_key(key_path)

    if did_doc_path:
        if issuer or algorithm:
            raise ValueError(
                "The --issuer and --alg flags may not be used together with a DID document."
            )

        with open(did_doc_path) as f:
            did_doc = json.load(f)
        return did.get_signer(key, did_doc, kid)
    elif x5c_path:
        ca_certs = _parse_x5c_file(x5c_path)

        return crypto.Signer(key, algorithm=algorithm, x5c=ca_certs)
    else:
        return crypto.Signer(key, issuer, kid, algorithm)


def sign_claims(
    claims_path: Path,
    key_path: Path,
    out_path: Path,
    did_doc_path: Optional[Path],
    issuer: Optional[str],
    content_type: str,
    algorithm: Optional[str],
    kid: Optional[str],
    feed: Optional[str],
    registration_info_args: List[RegistrationInfoArgument],
    x5c_path: Optional[str],
    akv_configuration_path: Optional[Path],
):
    # If a Key Vault configuration is provided, we sign with AKV
    if akv_configuration_path:
        if not x5c_path:
            raise ValueError(
                "The --x5c flag must be provided when signing with Azure Key Vault."
            )

        # Parse the AKV configuration file
        akv_sign_configuration_dict = json.loads(akv_configuration_path.read_text())

        # Parse the x5c file containing the x509 CA certificates
        ca_certs = _parse_x5c_file(x5c_path)

        kv_client = KeyVaultSignClient(akv_sign_configuration_dict)
        signed_claims = kv_client.cose_sign(
            claims_path.read_bytes(),
            {
                pycose.headers.ContentType: content_type,
                pycose.headers.X5chain: [crypto.cert_pem_to_der(x5) for x5 in ca_certs],
            },
        )
    else:
        signer = create_signer_from_arguments(
            key_path, did_doc_path, kid, issuer, algorithm, x5c_path
        )
        claims = claims_path.read_bytes()
        registration_info = {arg.name: arg.value() for arg in registration_info_args}

        signed_claims = crypto.sign_claimset(
            signer, claims, content_type, feed, registration_info
        )

    print(f"Writing {out_path}")
    out_path.write_bytes(signed_claims)


def cli(fn):
    parser = fn(description="Sign a claimset")
    parser.add_argument(
        "--claims", type=Path, required=True, help="Path to claims file"
    )
    parser.add_argument(
        "--key", type=Path, required=False, help="Path to PEM-encoded private key"
    )
    parser.add_argument(
        "--out",
        type=Path,
        required=True,
        help="Output path for signed claimset (must end in .cose)",
    )

    # Signing with a Key Vault certificate
    parser.add_argument(
        "--akv-configuration",
        type=Path,
        help="Path to an Azure key vault configuration file. The configuration is a JSON file containing the following fields: keyVaultName, certificateName, certificateVersion.",
    )

    # Signing with an existing DID document
    parser.add_argument("--did-doc", type=Path, help="Path to DID document")

    # Ad-hoc signing, without any on-disk document
    parser.add_argument("--issuer", help="Issuer stored in envelope header")
    parser.add_argument("--alg", help="Signing algorithm to use.")
    parser.add_argument("--x5c", help="Path to PEM-encoded certificate authority")

    parser.add_argument("--content-type", required=True, help="Content type of claims")
    parser.add_argument("--kid", help='Key ID ("kid" field) to use if multiple')
    parser.add_argument("--feed", help='Optional "feed" stored in envelope header')
    parser.add_argument(
        "--registration-info",
        metavar="[TYPE:]NAME=CONTENT",
        action="append",
        type=RegistrationInfoArgument,
        default=[],
        help="""
        Optional registration information to be stored in the envelope header.
        The flag may be specified multiple times, once per registration info entry.
        If content has the form `@file.txt`, the data will be read from the specified file instead.
        The type must be one of `text`, `bytes` or `int`. If not specified, the type defaults to text.
        """,
    )

    parser.set_defaults(
        func=lambda args: sign_claims(
            args.claims,
            args.key,
            args.out,
            args.did_doc,
            args.issuer,
            args.content_type,
            args.alg,
            args.kid,
            args.feed,
            args.registration_info,
            args.x5c,
            args.akv_configuration,
        )
    )

    return parser


if __name__ == "__main__":
    parser = cli(argparse.ArgumentParser)
    args = parser.parse_args()
    args.func(args)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import base64
import datetime
import hashlib
import json
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import uuid4

warnings.filterwarnings("ignore", category=Warning)

import cbor2
import jwt
import pycose.algorithms
import pycose.headers
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, padding, rsa
from cryptography.hazmat.primitives.asymmetric.ec import (
    EllipticCurve,
    EllipticCurvePrivateKey,
    EllipticCurvePublicKey,
    EllipticCurvePublicNumbers,
)
from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PrivateKey,
    Ed25519PublicKey,
)
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from cryptography.hazmat.primitives.asymmetric.utils import encode_dss_signature
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
    PublicFormat,
    load_pem_private_key,
    load_pem_public_key,
    load_ssh_private_key,
    load_ssh_public_key,
)
from cryptography.x509 import load_der_x509_certificate, load_pem_x509_certificate
from cryptography.x509.oid import NameOID
from pycose.keys.curves import P256, P384, P521, Ed25519
from pycose.keys.ec2 import EC2Key
from pycose.keys.keyparam import (
    EC2KpCurve,
    EC2KpD,
    EC2KpX,
    EC2KpY,
    OKPKpCurve,
    OKPKpD,
    OKPKpX,
    RSAKpD,
    RSAKpDP,
    RSAKpDQ,
    RSAKpE,
    RSAKpN,
    RSAKpP,
    RSAKpQ,
    RSAKpQInv,
)
from pycose.keys.okp import OKPKey
from pycose.keys.rsa import RSAKey
from pycose.messages import Sign1Message

RECOMMENDED_RSA_PUBLIC_EXPONENT = 65537

REGISTERED_EC_CURVES = {
    "P-256": P256,
    "P-384": P384,
    "P-521": P521,
}

Pem = str

COSE_HEADER_PARAM_ISSUER = 391
COSE_HEADER_PARAM_FEED = 392
COSE_HEADER_PARAM_REGISTRATION_INFO = 393
COSE_HEADER_PARAM_SCITT_RECEIPTS = 394

RegistrationInfoValue = Union[str, bytes, int]
RegistrationInfo = Dict[str, RegistrationInfoValue]


def generate_rsa_keypair(key_size: int) -> Tuple[Pem, Pem]:
    priv = rsa.generate_private_key(
        public_exponent=RECOMMENDED_RSA_PUBLIC_EXPONENT,
        key_size=key_size,
    )
    pub = priv.public_key()
    priv_pem = priv.private_bytes(
        Encoding.PEM, PrivateFormat.PKCS8, NoEncryption()
    ).decode("ascii")
    pub_pem = pub.public_bytes(Encoding.PEM, PublicFormat.SubjectPublicKeyInfo).decode(
        "ascii"
    )
    return priv_pem, pub_pem


def generate_ec_keypair(curve: str) -> Tuple[Pem, Pem]:
    if curve not in REGISTERED_EC_CURVES:
        raise NotImplementedError(f"Unsupported curve: {curve}")
    curve_obj = REGISTERED_EC_CURVES[curve].curve_obj
    assert isinstance(curve_obj, EllipticCurve)
    priv = ec.generate_private_key(curve=curve_obj)
    pub = priv.public_key()
    priv_pem = priv.private_bytes(
        Encoding.PEM, PrivateFormat.PKCS8, NoEncryption()
    ).decode("ascii")
    pub_pem = pub.public_bytes(Encoding.PEM, PublicFormat.SubjectPublicKeyInfo).decode(
        "ascii"
    )
    return priv_pem, pub_pem


def generate_ed25519_keypair() -> Tuple[Pem, Pem]:
    key = Ed25519PrivateKey.generate()

    priv_pem = key.private_bytes(
        Encoding.PEM, PrivateFormat.PKCS8, NoEncryption()
    ).decode("ascii")
    pub_pem = (
        key.public_key()
        .public_bytes(Encoding.PEM, PublicFormat.SubjectPublicKeyInfo)
        .decode("ascii")
    )
    return priv_pem, pub_pem


def generate_keypair(
    kty: str,
    *,
    rsa_key_size: Optional[int] = None,
    ec_curve: Optional[str] = None,
) -> Tuple[str, str]:
    if kty == "rsa":
        return generate_rsa_keypair(rsa_key_size or 2048)
    elif kty == "ec":
        return generate_ec_keypair(ec_curve or "P-256")
    elif kty == "ed25519":
        return generate_ed25519_keypair()
    else:
        raise ValueError(f"Unsupported key type: {kty}")


def is_ssh_private_key(pem: str):
    return pem.startswith("-----BEGIN OPENSSH")


def ssh_private_key_to_pem(pem: str) -> Pem:
    priv = load_ssh_private_key(pem.encode("ascii"), None)
    pem = priv.private_bytes(Encoding.PEM, PrivateFormat.PKCS8, NoEncryption()).decode(
        "ascii"
    )
    return pem


def ssh_public_key_to_pem(ssh_pub_key: str) -> Pem:
    pub = load_ssh_public_key(ssh_pub_key.encode("ascii"))
    pem = pub.public_bytes(Encoding.PEM, PublicFormat.SubjectPublicKeyInfo).decode(
        "ascii"
    )
    return pem


def generate_cert(
    private_key_pem: Pem,
    *,
    issuer: Optional[Tuple[Pem, Pem]] = None,
    ca: bool = False,
    cn: Optional[str] = None,
):
    if not cn:
        cn = str(uuid4())
    subject_priv = load_pem_private_key(private_key_pem.encode("ascii"), None)
    assert isinstance(
        subject_priv, (RSAPrivateKey, EllipticCurvePrivateKey, Ed25519PrivateKey)
    )

    subject_pub_key = subject_priv.public_key()

    subject_name = x509.Name(
        [
            x509.NameAttribute(NameOID.COMMON_NAME, cn),
        ]
    )

    if isinstance(subject_priv, Ed25519PrivateKey):
        hash_alg = None
    else:
        hash_alg = hashes.SHA256()

    if issuer is not None:
        issuer_name = load_pem_x509_certificate(issuer[0].encode("ascii")).subject
        issuer_key = load_pem_private_key(
            issuer[1].encode("ascii"),
            None,
        )
        assert isinstance(
            issuer_key, (RSAPrivateKey, EllipticCurvePrivateKey, Ed25519PrivateKey)
        )
    else:
        issuer_name = subject_name
        issuer_key = subject_priv

    cert = (
        x509.CertificateBuilder()
        .subject_name(subject_name)
        .issuer_name(issuer_name)
        .public_key(subject_pub_key)
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.datetime.utcnow())
        .not_valid_after(datetime.datetime.utcnow() + datetime.timedelta(days=10))
        .add_extension(
            x509.KeyUsage(
                digital_signature=not ca,
                content_commitment=False,
                key_encipherment=False,
                data_encipherment=False,
                key_agreement=False,
                key_cert_sign=ca,
                crl_sign=False,
                encipher_only=False,
                decipher_only=False,
            ),
            critical=True,
        )
        .add_extension(x509.BasicConstraints(ca=ca, path_length=None), critical=True)
        .sign(issuer_key, hash_alg)
    )
    return cert.public_bytes(Encoding.PEM).decode("ascii")


def get_priv_key_type(priv_pem: str) -> str:
    key = load_pem_private_key(priv_pem.encode("ascii"), None)
    if isinstance(key, RSAPrivateKey):
        return "rsa"
    elif isinstance(key, EllipticCurvePrivateKey):
        return "ec"
    elif isinstance(key, Ed25519PrivateKey):
        return "ed25519"
    raise NotImplementedError("unsupported key type")


def get_pub_key_type(pub_pem: str) -> str:
    key = load_pem_public_key(pub_pem.encode("ascii"))
    if isinstance(key, RSAPublicKey):
        return "rsa"
    elif isinstance(key, EllipticCurvePublicKey):
        return "ec"
    elif isinstance(key, Ed25519PublicKey):
        return "ed25519"
    raise NotImplementedError("unsupported key type")


def get_cert_info(pem: str) -> dict:
    cert = load_pem_x509_certificate(pem.encode("ascii"))
    cn = cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value
    return {"cn": cn}


def get_cert_public_key(pem: Pem) -> Pem:
    cert = load_pem_x509_certificate(pem.encode("ascii"))
    key = cert.public_key()
    return key.public_bytes(Encoding.PEM, PublicFormat.SubjectPublicKeyInfo).decode(
        "ascii"
    )


def get_cert_fingerprint(pem: Pem) -> str:
    cert = load_pem_x509_certificate(pem.encode("ascii"))
    return cert.fingerprint(hashes.SHA256()).hex()


def get_public_key_fingerprint(pem: Pem) -> str:
    pub_key = load_pem_public_key(pem.encode("ascii"))
    der = pub_key.public_bytes(Encoding.DER, PublicFormat.SubjectPublicKeyInfo)
    return hashlib.sha256(der).hexdigest()


def pub_key_pem_to_der(pem: Pem) -> bytes:
    pub_key = load_pem_public_key(pem.encode("ascii"))
    return pub_key.public_bytes(Encoding.DER, PublicFormat.SubjectPublicKeyInfo)


def pub_key_pem_to_ssh(pem: Pem) -> str:
    pub_key = load_pem_public_key(pem.encode("ascii"))
    return pub_key.public_bytes(Encoding.OpenSSH, PublicFormat.OpenSSH).decode("ascii")


def private_key_to_public(pem: Pem) -> Pem:
    private_key = load_pem_private_key(pem.encode("ascii"), None)
    return (
        private_key.public_key()
        .public_bytes(Encoding.PEM, PublicFormat.SubjectPublicKeyInfo)
        .decode("ascii")
    )


def private_key_pem_to_ssh(pem: Pem) -> str:
    private_key = load_pem_private_key(pem.encode("ascii"), None)
    return private_key.private_bytes(
        Encoding.PEM, PrivateFormat.OpenSSH, NoEncryption()
    ).decode("ascii")


def cert_pem_to_der(pem: Pem) -> bytes:
    cert = load_pem_x509_certificate(pem.encode("ascii"))
    return cert.public_bytes(Encoding.DER)


def cert_der_to_pem(der: bytes) -> str:
    cert = load_der_x509_certificate(der)
    return cert.public_bytes(Encoding.PEM).decode("ascii")


def create_jwt(body_claims: dict, key_priv_pem: str, key_id: str) -> str:
    return jwt.encode(
        body_claims, key_priv_pem, algorithm="RS256", headers={"kid": key_id}
    )


def create_jwks(cert_pem: str, kid: str):
    der_b64 = base64.b64encode(cert_pem_to_der(cert_pem)).decode("ascii")
    return {"keys": [{"kty": "RSA", "kid": kid, "x5c": [der_b64]}]}


def default_algorithm_for_key(key) -> str:
    """
    Get the default algorithm for a given key, based on its
    type and parameters.
    """

    if isinstance(key, (RSAPublicKey, RSAPrivateKey)):
        return "PS256"

    elif isinstance(key, (EllipticCurvePublicKey, EllipticCurvePrivateKey)):
        if isinstance(key.curve, ec.SECP256R1):
            return "ES256"
        elif isinstance(key.curve, ec.SECP384R1):
            return "ES384"
        elif isinstance(key.curve, ec.SECP521R1):
            return "ES512"
        else:
            raise NotImplementedError("unsupported curve")

    elif isinstance(key, (Ed25519PublicKey, Ed25519PrivateKey)):
        return "EdDSA"
    else:
        raise NotImplementedError(f"unsupported key type: {type(key)}")


def default_algorithm_for_private_key(key_pem: Pem) -> str:
    key = load_pem_private_key(key_pem.encode("ascii"), None)
    return default_algorithm_for_key(key)


def cose_header_to_jws_header(cose_header: dict) -> dict:
    jws_header = {}
    transforms = {
        # registered
        "ALG": ("alg", lambda v: v.fullname),
        "KID": ("kid", lambda v: v.decode("utf-8")),
        "CONTENT_TYPE": ("cty", None),
        "X5_CHAIN": ("x5c", lambda v: [base64.b64encode(c).decode() for c in v]),
    }
    for k, v in cose_header.items():
        try:
            k = k.fullname
        except AttributeError:
            pass
        if k in transforms:
            k, f = transforms[k]
            if f is not None:
                v = f(v)
        jws_header[k] = v
    return jws_header


def parse_cose_sign1(buf: bytes) -> Tuple[dict, bytes]:
    msg = Sign1Message.decode(buf)
    header = cose_header_to_jws_header(msg.phdr)
    payload = msg.payload
    assert payload, "Payload is null"
    return header, payload


def pretty_cose_sign1(buf: bytes) -> str:
    header, payload = parse_cose_sign1(buf)
    try:
        claims = json.loads(payload)
    except json.JSONDecodeError:
        claims = f"<{len(payload)} bytes>"
    try:
        return (
            "COSE_Sign1(\nheader="
            + json.dumps(header, indent=2)
            + "\npayload="
            + json.dumps(claims, indent=2)
            + "\n<signature>)"
        )
    except TypeError:
        print(header)
        raise


# temporary, from https://github.com/BrianSipos/pycose/blob/rsa_keys_algs/cose/keys/rsa.py
# until https://github.com/TimothyClaeys/pycose/issues/44 is implemented
def from_cryptography_rsakey_obj(ext_key: Union[RSAPrivateKey, RSAPublicKey]) -> RSAKey:
    """
    Returns an initialized COSE Key object of type RSAKey.
    :param ext_key: Python cryptography key.
    :return: an initialized RSA key
    """

    def to_bstr(dec):
        blen = (dec.bit_length() + 7) // 8
        return dec.to_bytes(blen, byteorder="big")

    if hasattr(ext_key, "private_numbers"):
        priv_nums = ext_key.private_numbers()
        pub_nums = priv_nums.public_numbers
    else:
        priv_nums = None
        pub_nums = ext_key.public_numbers()

    cose_key = {}
    if pub_nums:
        cose_key.update(
            {
                RSAKpE: to_bstr(pub_nums.e),
                RSAKpN: to_bstr(pub_nums.n),
            }
        )
    if priv_nums:
        cose_key.update(
            {
                RSAKpD: to_bstr(priv_nums.d),
                RSAKpP: to_bstr(priv_nums.p),
                RSAKpQ: to_bstr(priv_nums.q),
                RSAKpDP: to_bstr(priv_nums.dmp1),
                RSAKpDQ: to_bstr(priv_nums.dmq1),
                RSAKpQInv: to_bstr(priv_nums.iqmp),
            }
        )
    return RSAKey.from_dict(cose_key)


def from_cryptography_eckey_obj(
    ext_key: Union[EllipticCurvePrivateKey, EllipticCurvePublicKey]
) -> EC2Key:
    """
    Returns an initialized COSE Key object of type EC2Key.
    :param ext_key: Python cryptography key.
    :return: an initialized EC key
    """
    if hasattr(ext_key, "private_numbers"):
        priv_nums = ext_key.private_numbers()
        pub_nums = priv_nums.public_numbers
    else:
        priv_nums = None
        pub_nums = ext_key.public_numbers()

    # Create map of cryptography curves to cose curves. E.g. {ec.SECP256R1: P256, ...}
    registered_crvs = {
        type(crv.curve_obj): crv for crv in REGISTERED_EC_CURVES.values()
    }
    if type(pub_nums.curve) not in registered_crvs:
        raise ValueError(f"Unsupported EC Curve: {type(pub_nums.curve)}")
    curve = registered_crvs[type(pub_nums.curve)]

    cose_key = {}
    if pub_nums:
        cose_key.update(
            {
                EC2KpCurve: curve,
                EC2KpX: pub_nums.x.to_bytes(curve.size, "big"),
                EC2KpY: pub_nums.y.to_bytes(curve.size, "big"),
            }
        )
    if priv_nums:
        cose_key.update(
            {
                EC2KpD: priv_nums.private_value.to_bytes(curve.size, "big"),
            }
        )
    return EC2Key.from_dict(cose_key)


def from_cryptography_ed25519key_obj(
    ext_key: Union[Ed25519PrivateKey, Ed25519PublicKey]
) -> OKPKey:
    """
    Returns an initialized COSE Key object of type OKPKey.
    :param ext_key: Python cryptography key.
    :return: an initialized OKP key
    """
    if isinstance(ext_key, Ed25519PrivateKey):
        priv_bytes = ext_key.private_bytes(
            encoding=Encoding.Raw,
            format=PrivateFormat.Raw,
            encryption_algorithm=NoEncryption(),
        )
        pub_bytes = ext_key.public_key().public_bytes(
            encoding=Encoding.Raw, format=PublicFormat.Raw
        )
    else:
        assert isinstance(ext_key, Ed25519PublicKey)
        priv_bytes = None
        pub_bytes = ext_key.public_bytes(encoding=Encoding.Raw, format=PublicFormat.Raw)

    curve = Ed25519

    cose_key = {}
    cose_key.update(
        {
            OKPKpCurve: curve,
            OKPKpX: pub_bytes,
        }
    )
    if priv_bytes:
        cose_key.update(
            {
                OKPKpD: priv_bytes,
            }
        )
    return OKPKey.from_dict(cose_key)


def cose_private_key_from_pem(pem: Pem):
    key = load_pem_private_key(pem.encode("ascii"), None)
    if isinstance(key, RSAPrivateKey):
        return from_cryptography_rsakey_obj(key)
    elif isinstance(key, EllipticCurvePrivateKey):
        return from_cryptography_eckey_obj(key)
    elif isinstance(key, Ed25519PrivateKey):
        return from_cryptography_ed25519key_obj(key)
    raise NotImplementedError("unsupported key type")


def b64url(b: bytes) -> str:
    return base64.b64encode(b, altchars=b"-_").decode("ascii")


def sha256_file(path: Path) -> str:
    content = open(path, "rb").read()
    sha256 = hashlib.sha256(content).hexdigest()
    return sha256


def embed_receipt_in_cose(buf: bytes, receipt: bytes) -> bytes:
    # Need to parse the receipt to avoid wrapping it in a bstr.
    parsed_receipt = cbor2.loads(receipt)

    outer = cbor2.loads(buf)
    if hasattr(outer, "tag"):
        assert outer.tag == 18  # COSE_Sign1
        val = outer.value  # type: ignore[attr-defined]
    else:
        val = outer
    [_, uhdr, _, _] = val
    key = COSE_HEADER_PARAM_SCITT_RECEIPTS
    if key not in uhdr:
        uhdr[key] = []
    uhdr[key].append(parsed_receipt)
    return cbor2.dumps(outer)


def load_private_key(key_path: Path) -> Pem:
    with open(key_path) as f:
        key_priv_pem = f.read()
    if is_ssh_private_key(key_priv_pem):
        key_priv_pem = ssh_private_key_to_pem(key_priv_pem)
    return key_priv_pem


def jwk_from_public_key(
    pem: Pem,
    alg: Optional[str] = None,
    kid: Optional[str] = None,
):
    pub_key = load_pem_public_key(pem.encode("ascii"))
    if isinstance(pub_key, RSAPublicKey):
        pub_nums = pub_key.public_numbers()

        def encode_pub_num_jwk(dec):
            blen = (dec.bit_length() + 7) // 8
            b = dec.to_bytes(blen, byteorder="big")
            return base64.urlsafe_b64encode(b).decode("ascii")

        jwk = {
            "kty": "RSA",
            "n": encode_pub_num_jwk(pub_nums.n),
            "e": encode_pub_num_jwk(pub_nums.e),
        }

    elif isinstance(pub_key, Ed25519PublicKey):
        x = pub_key.public_bytes(Encoding.Raw, PublicFormat.Raw)
        jwk = {
            "kty": "OKP",
            "crv": "Ed25519",
            "x": base64.urlsafe_b64encode(x).decode("ascii"),
        }

    elif isinstance(pub_key, EllipticCurvePublicKey):
        pub_numbers = pub_key.public_numbers()
        curve = pub_numbers.curve
        # Create map of curves to names. E.g. {ec.SECP256R1: "P-256", ...}
        registered_crvs = {
            type(crv.curve_obj): name for name, crv in REGISTERED_EC_CURVES.items()
        }
        if type(curve) not in registered_crvs:
            raise ValueError(f"Unsupported EC Curve: {curve}")
        crv_name = registered_crvs[type(curve)]
        x = pub_numbers.x.to_bytes(REGISTERED_EC_CURVES[crv_name].size, "big")
        y = pub_numbers.y.to_bytes(REGISTERED_EC_CURVES[crv_name].size, "big")

        jwk = {
            "kty": "EC",
            "crv": crv_name,
            "x": base64.urlsafe_b64encode(x).decode("ascii"),
            "y": base64.urlsafe_b64encode(y).decode("ascii"),
        }
    else:
        raise ValueError("unsupported key type")

    if kid is not None:
        jwk["kid"] = kid

    if alg is None:
        alg = default_algorithm_for_key(pub_key)
    jwk["alg"] = alg

    return jwk


@dataclass(init=False)
class Signer:
    private_key: Pem
    issuer: Optional[str]
    kid: Optional[str]
    algorithm: str
    x5c: Optional[List[Pem]]

    def __init__(
        self,
        private_key: Pem,
        issuer: Optional[str] = None,
        kid: Optional[str] = None,
        algorithm: Optional[str] = None,
        x5c: Optional[List[Pem]] = None,
    ):
        """
        If no algorithm is specified, a sensible default is inferred from the private key.
        """

        self.private_key = private_key
        self.issuer = issuer
        self.kid = kid
        self.algorithm = algorithm or default_algorithm_for_private_key(private_key)
        self.x5c = x5c


def sign_claimset(
    signer: Signer,
    claims: bytes,
    content_type: str,
    feed: Optional[str] = None,
    registration_info: RegistrationInfo = {},
) -> bytes:
    headers: dict = {}
    headers[pycose.headers.Algorithm] = signer.algorithm
    headers[pycose.headers.ContentType] = content_type

    if signer.x5c is not None:
        headers[pycose.headers.X5chain] = [cert_pem_to_der(x5) for x5 in signer.x5c]
    if signer.kid is not None:
        headers[pycose.headers.KID] = signer.kid.encode("utf-8")
    if signer.issuer is not None:
        headers[COSE_HEADER_PARAM_ISSUER] = signer.issuer
    if feed is not None:
        headers[COSE_HEADER_PARAM_FEED] = feed
    if registration_info:
        headers[COSE_HEADER_PARAM_REGISTRATION_INFO] = registration_info

    msg = Sign1Message(phdr=headers, payload=claims)
    msg.key = cose_private_key_from_pem(signer.private_key)
    return msg.encode(tag=True)


def sign_json_claimset(
    signer: Signer,
    claims: Any,
    content_type: str = "application/vnd.dummy+json",
    feed: Optional[str] = None,
) -> bytes:
    return sign_claimset(
        signer, json.dumps(claims).encode("ascii"), content_type=content_type, feed=feed
    )


def decode_p1363_signature(signature: bytes) -> Tuple[int, int]:
    """
    Decode an ECDSA signature from its IEEE P1363 encoding into its r and s
    components. The two integers are padded to the curve size and concatenated.

    This is the format used throughout the COSE/JOSE ecosystem.
    """
    # The two components are padded to the same size, so we can find the size
    # of each one by taking half the size of the signature.
    assert len(signature) % 2 == 0
    mid = len(signature) // 2
    r = int.from_bytes(signature[:mid], "big")
    s = int.from_bytes(signature[mid:], "big")
    return r, s


def convert_p1363_signature_to_dss(signature: bytes) -> bytes:
    """
    Convert an ECDSA signature from its IEEE P1363 encoding to an ASN1/DER
    encoding.

    The former is the format used throughout the COSE/JOSE ecosystem. The
    latter is used by OpenSSL and cryptography, as well as the CCF python
    module.

    """
    r, s = decode_p1363_signature(signature)
    return encode_dss_signature(r, s)


def convert_jwk_to_pem(jwk: dict) -> Pem:
    if jwk.get("kty") == "EC":
        x = int.from_bytes(base64.urlsafe_b64decode(jwk["x"]), "big")
        y = int.from_bytes(base64.urlsafe_b64decode(jwk["y"]), "big")
        crv = REGISTERED_EC_CURVES[jwk["crv"]].curve_obj
        assert isinstance(crv, EllipticCurve)
        key = EllipticCurvePublicNumbers(x, y, crv).public_key()
    else:
        raise NotImplementedError("Unsupported JWK type")

    return key.public_bytes(Encoding.PEM, PublicFormat.SubjectPublicKeyInfo).decode(
        "ascii"
    )


def decrypt_recovery_share(key_pem: Pem, encrypted_share: bytes) -> bytes:
    """
    Decrypt a CCF recovery share, using the member's encryption private key.

    https://microsoft.github.io/CCF/release/3.x/governance/accept_recovery.html
    """
    key = load_pem_private_key(key_pem.encode("ascii"), None)
    assert isinstance(key, RSAPrivateKey)
    return key.decrypt(
        encrypted_share,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA256()),
            algorithm=hashes.SHA256(),
            label=None,
        ),
    )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import argparse

from . import (
    create_did_web,
    embed_receipt_in_cose,
    governance,
    prefix_tree,
    pretty_receipt,
    retrieve_signed_claims,
    sign_claims,
    submit_signed_claims,
    upload_did_web_doc_to_github,
    validate_cose,
)

COMMANDS = [
    ("create-did-web", create_did_web),
    ("upload-did-web-github", upload_did_web_doc_to_github),
    ("sign", sign_claims),
    ("submit", submit_signed_claims),
    ("retrieve", retrieve_signed_claims),
    ("pretty-receipt", pretty_receipt),
    ("embed-receipt", embed_receipt_in_cose),
    ("validate", validate_cose),
    ("governance", governance),
    ("prefix-tree", prefix_tree),
]


def main(argv=None):
    parser = argparse.ArgumentParser()
    sub = parser.add_subparsers()
    for name, module in COMMANDS:
        getattr(module, "cli")(lambda *args, **kw: sub.add_parser(name, *args, **kw))
    args = parser.parse_args(argv)
    if not hasattr(args, "func"):
        parser.print_usage()
    else:
        args.func(args)


if __name__ == "__main__":
    main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import argparse
from pathlib import Path
from typing import Optional

from .. import crypto


def embed_receipt_in_cose(
    cose_path: Path, receipt_path: Path, out_path: Optional[Path]
):
    with open(cose_path, "rb") as f:
        cose = f.read()
    with open(receipt_path, "rb") as f:
        receipt = f.read()
    cose = crypto.embed_receipt_in_cose(cose, receipt)
    if out_path is None:
        out_path = cose_path
    with open(out_path, "wb") as f:
        f.write(cose)
    print(f"Receipt embedded: {out_path}")


def cli(fn):
    parser = fn(description="Embed a receipt into a COSE_Sign1 document")
    parser.add_argument("cose", type=Path, help="Path to COSE file")
    parser.add_argument("--receipt", required=True, type=Path, help="Path to receipt")
    parser.add_argument(
        "--out", type=Path, help="Optional output path, otherwise input file"
    )

    def cmd(args):
        embed_receipt_in_cose(args.cose, args.receipt, args.out)

    parser.set_defaults(func=cmd)

    return parser


if __name__ == "__main__":
    parser = cli(argparse.ArgumentParser)
    args = parser.parse_args()
    args.func(args)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import json
from pathlib import Path
from typing import Optional

from pycose.messages import CoseMessage

from .. import prefix_tree
from ..client import Client
from ..crypto import COSE_HEADER_PARAM_FEED, COSE_HEADER_PARAM_ISSUER
from ..verify import StaticTrustStore
from .client_arguments import add_client_arguments, create_client


def prefix_tree_flush(client: Client):
    data = client.prefix_tree.flush()
    print(json.dumps(data, indent=2))


def prefix_tree_debug(client: Client):
    data = client.prefix_tree.debug()
    print(json.dumps(data, indent=2))


def prefix_tree_get_receipt(
    client: Client,
    claim_path: Optional[Path],
    issuer: Optional[str],
    feed: Optional[str],
    output: Optional[Path],
    service_trust_store_path: Path,
):
    if claim_path:
        claim = CoseMessage.decode(claim_path.read_bytes())

        issuer = claim.phdr[COSE_HEADER_PARAM_ISSUER]
        feed = claim.phdr[COSE_HEADER_PARAM_FEED]
    elif not issuer or not feed:
        raise ValueError("Either a claim or an issuer and feed must be specified.")

    receipt_data = client.prefix_tree.get_read_receipt(issuer, feed, decode=False)
    if output:
        with open(output, "wb") as f:
            f.write(receipt_data)

    receipt = prefix_tree.ReadReceipt.decode(receipt_data)
    print(json.dumps(receipt.as_dict(), indent=2))

    if claim_path and service_trust_store_path:
        service_trust_store = StaticTrustStore.load(service_trust_store_path)
        service_params = service_trust_store.lookup(receipt.tree_headers)
        receipt.verify(claim, service_params)


def cli(fn):
    parser = fn(description="Manipulate the prefix tree")
    sub = parser.add_subparsers(help="Action to execute", required=True)

    p = sub.add_parser(
        "debug", description="Display an internal representation of the prefix tree"
    )
    add_client_arguments(p)
    p.set_defaults(func=lambda args: prefix_tree_debug(create_client(args)))

    p = sub.add_parser("flush", description="Flush pending entries to the prefix tree")
    add_client_arguments(p)
    p.set_defaults(func=lambda args: prefix_tree_flush(create_client(args)))

    p = sub.add_parser(
        "receipt",
        description="Fetch a read receipt, either by issuer and feed or for an existing claim.",
    )
    add_client_arguments(p)

    group = p.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--claim",
        type=Path,
        help="Look up a read-receipt corresponding to the given claim.",
    )
    group.add_argument("--issuer", help="Look up a read-receipt by issuer and feed.")
    p.add_argument("--feed")
    p.add_argument("--output", type=Path, help="Output path to receipt file")
    p.add_argument(
        "--service-trust-store",
        type=Path,
        help="Folder containing JSON parameter files of SCITT services to trust",
    )
    p.set_defaults(
        func=lambda args: prefix_tree_get_receipt(
            create_client(args),
            args.claim,
            args.issuer,
            args.feed,
            args.output,
            args.service_trust_store,
        )
    )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import argparse
import json
from pathlib import Path

from pyscitt.key_vault_sign_client import KeyVaultSignClient
from pyscitt.local_key_sign_client import LocalKeySignClient

from ..client import Client

CCF_URL_DEFAULT = "https://127.0.0.1:8000"
CCF_SANDBOX_WORKSPACE = Path("workspace")
CCF_MEMBER_KEY_DEFAULT = CCF_SANDBOX_WORKSPACE / "member0_privk.pem"
CCF_MEMBER_CERT_DEFAULT = CCF_SANDBOX_WORKSPACE / "member0_cert.pem"


def add_client_arguments(
    parser: argparse.ArgumentParser,
    *,
    with_member_auth: bool = False,
    with_auth_token: bool = False,
    development_only: bool = False,
):
    """
    Add command-line arguments to an argparse parser, such that a client instance can be configured.

    with_member_auth:
        If True, required flag arguments for a member certificate and private key will be added
        to the parser.

    with_auth_token:
        If True, an optional flag argument allowing a bearer token to be specified will be added
        to the parser.

    development_only:
        If True, the server's TLS certificate is never verified, and no --development flag is added
        to the parser.
    """

    parser.add_argument("--url", help="CCF service URL", default=CCF_URL_DEFAULT)
    parser.add_argument(
        "--cacert",
        help="Path to certificate file (must be in PEM format) to verify the CCF service",
        default=None,
    )
    if development_only:
        # We always add a hidden --development flag that defaults to True.
        # This helps provide a uniform interface for scripts and tests.
        parser.add_argument(
            "-k",
            "--development",
            action="store_true",
            help=argparse.SUPPRESS,
            default=True,
        )
    else:
        parser.add_argument(
            "-k",
            "--development",
            action="store_true",
            help="Do not verify the self-signed network certificate",
        )

    if with_member_auth:
        parser.add_argument(
            "--member-key",
            default=CCF_MEMBER_KEY_DEFAULT,
            type=Path,
            help="Path to CCF member key file",
        )
        parser.add_argument(
            "--member-cert",
            default=CCF_MEMBER_CERT_DEFAULT,
            type=Path,
            help="Path to CCF member certificate file",
        )
        parser.add_argument(
            "--akv-configuration",
            type=Path,
            help="Path to an Azure key vault configuration file. The configuration is a JSON file containing the following fields: keyVaultName, certificateName, certificateVersion.",
        )

    if with_auth_token:
        parser.add_argument("--auth-token", help="JWT bearer token")


def create_client(args: argparse.Namespace):
    """
    Create a client instance from the result of parsing command-line arguments.

    This assumes the argument parser was configured with the `add_client_arguments` function.
    """
    kwargs = {
        "url": args.url,
        "cacert": args.cacert,
        "development": args.development,
    }

    if "akv_configuration" in args and args.akv_configuration:
        akv_configuration = args.akv_configuration.read_text()
        akv_sign_configuration_dict = json.loads(akv_configuration)
        kwargs["member_auth"] = KeyVaultSignClient(akv_sign_configuration_dict)
    elif "member_cert" in args:
        cert = args.member_cert.read_text()
        key = args.member_key.read_text()
        kwargs["member_auth"] = LocalKeySignClient(cert, key)

    if "auth_token" in args:
        kwargs["auth_token"] = args.auth_token

    return Client(**kwargs)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .client import Client

import base64
import json
from dataclasses import dataclass
from typing import Optional, Union

from . import crypto


@dataclass
class SubmittedProposal:
    id: str
    state: str

    @property
    def is_open(self):
        return self.state == "Open"

    @property
    def is_accepted(self):
        return self.state == "Accepted"


class ProposalNotAccepted(Exception): ...


class GovernanceClient:
    client: "Client"

    def __init__(self, client: "Client"):
        self.client = client

    def propose(
        self,
        proposal: Union[dict, bytes],
        *,
        vote: bool = True,
        must_pass: bool = False,
    ) -> SubmittedProposal:
        """
        Submit a proposal to the SCITT CCF instance.

        Depending on the instance's constitution, proposals may not pass immediately and may need
        more members to cast a favourable ballot. The caller should either check this method's
        return value or use the `must_pass` argument before assuming the proposal is effective.

        vote:
            If True and the proposal isn't automatically accepted, an unconditional vote in its
            favour is cast.

        must_pass:
            If True and the proposal fails to pass even after a ballot in its favour is cast, an
            exception is raised.
        """
        assert vote or (not must_pass)

        if isinstance(proposal, dict):
            proposal = json.dumps(proposal).encode("ascii")

        response = self.client.post(
            "/gov/proposals", content=proposal, sign_request=True
        )
        out = response.json()
        result = SubmittedProposal(out["proposal_id"], out["state"])
        if vote and result.is_open:
            result = self.vote(result.id, must_pass=must_pass)

        return result

    def vote(
        self, proposal_id: str, ballot: Optional[str] = None, must_pass: bool = False
    ) -> SubmittedProposal:
        if ballot is None:
            ballot = """
                export function vote (rawProposal, proposerId) {
                    return true;
                }
            """
        response = self.client.post(
            f"/gov/proposals/{proposal_id}/ballots",
            sign_request=True,
            json={
                "ballot": ballot,
            },
        )

        out = response.json()
        result = SubmittedProposal(out["proposal_id"], out["state"])

        if must_pass and not result.is_accepted:
            raise ProposalNotAccepted(
                f"Proposal {result.id} was not accepted: {result.state}"
            )

        return result

    def get_recovery_share(self, key: Optional[crypto.Pem] = None) -> bytes:
        encoded_share = self.client.get(
            "/gov/recovery_share", sign_request=True
        ).json()["encrypted_share"]
        encrypted_share = base64.b64decode(encoded_share)

        if key is not None:
            return crypto.decrypt_recovery_share(key, encrypted_share)
        else:
            return encrypted_share

    def submit_recovery_share(self, share: bytes) -> None:
        self.client.post(
            "/gov/recovery_share",
            sign_request=True,
            json={"share": base64.b64encode(share).decode("ascii")},
        )

    def recover_service(self, encryption_private_key: crypto.Pem) -> None:
        """
        Perform a disaster recovery of the service. This assumes the service
        has a single member.

        https://microsoft.github.io/CCF/release/3.x/operations/recovery.html
        https://microsoft.github.io/CCF/release/3.x/governance/accept_recovery.html
        """
        recovery_share = self.get_recovery_share(encryption_private_key)
        previous_service_identity = self.client.get_previous_service_identity()
        next_service_identity = self.client.get_service_certificate()

        proposal = transition_service_to_open_proposal(
            next_service_identity,
            previous_service_identity,
        )
        self.propose(proposal, must_pass=True)
        self.submit_recovery_share(recovery_share)
        self.client.wait_for_network_open()

    def activate_member(self):
        r = self.client.post("/gov/ack/update_state_digest", sign_request=True)
        self.client.post("/gov/ack", content=r.content, sign_request=True)


def set_scitt_configuration_proposal(configuration: dict) -> dict:
    return {
        "actions": [
            {
                "name": "set_scitt_configuration",
                "args": {
                    "configuration": configuration,
                },
            }
        ]
    }


def set_ca_bundle_proposal(name: str, bundle: str) -> dict:
    return {
        "actions": [
            {
                "name": "set_ca_cert_bundle",
                "args": {"name": name, "cert_bundle": bundle},
            }
        ]
    }


def transition_service_to_open_proposal(
    next_service_identity: str, previous_service_identity: Optional[str] = None
) -> dict:
    args = {"next_service_identity": next_service_identity}

    if previous_service_identity:
        args["previous_service_identity"] = previous_service_identity

    return {
        "actions": [
            {
                "name": "transition_service_to_open",
                "args": args,
            }
        ]
    }


def set_constitution_proposal(constitution: str):
    return {
        "actions": [
            {
                "name": "set_constitution",
                "args": {
                    "constitution": constitution,
                },
            }
        ]
    }



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import base64
import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, Union

import cbor2
import pycose
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.x509 import load_pem_x509_certificate
from pycose.keys.ec2 import EC2Key
from pycose.keys.rsa import RSAKey
from pycose.messages import Sign1Message

from . import crypto, did
from .crypto import COSE_HEADER_PARAM_ISSUER, COSE_HEADER_PARAM_SCITT_RECEIPTS
from .receipt import Receipt


@dataclass
class ServiceParameters:
    tree_algorithm: str
    signature_algorithm: str
    certificate: bytes
    service_id: Optional[str] = None

    @staticmethod
    def from_dict(data) -> "ServiceParameters":
        """
        Decode service parameters as returned by the /parameters endpoint.
        """
        return ServiceParameters(
            tree_algorithm=data["treeAlgorithm"],
            signature_algorithm=data["signatureAlgorithm"],
            certificate=base64.b64decode(data["serviceCertificate"]),
            service_id=data["serviceId"],
        )

    def as_dict(self) -> Dict[str, str]:
        result = {
            "treeAlgorithm": self.tree_algorithm,
            "signatureAlgorithm": self.signature_algorithm,
            "serviceCertificate": base64.b64encode(self.certificate).decode("ascii"),
        }
        if self.service_id is not None:
            result["serviceId"] = self.service_id
        return result


class TrustStore(ABC):
    @abstractmethod
    def lookup(self, phdr) -> ServiceParameters:
        """
        Look up a service's parameters based on the protected headers from a
        receipt. Raises an exception if not matching service is found.
        """
        raise NotImplementedError()


def verify_cose_sign1(buf: bytes, cert_pem: str):
    cert = load_pem_x509_certificate(cert_pem.encode("ascii"))
    key = cert.public_key()

    cose_key: Union[EC2Key, RSAKey]
    if isinstance(key, RSAPublicKey):
        cose_key = crypto.from_cryptography_rsakey_obj(key)
    elif isinstance(key, EllipticCurvePublicKey):
        cose_key = crypto.from_cryptography_eckey_obj(key)
    else:
        raise NotImplementedError("unsupported key type")

    msg = Sign1Message.decode(buf)
    msg.key = cose_key
    if not msg.verify_signature():
        raise ValueError("signature is invalid")


def verify_receipt(
    buf: bytes,
    service_trust_store: TrustStore,
    receipt: Union[Receipt, bytes, None] = None,
) -> None:
    msg = Sign1Message.decode(buf)

    if isinstance(receipt, Receipt):
        decoded_receipt = receipt
    else:
        if receipt is None:
            parsed_receipts = msg.uhdr[COSE_HEADER_PARAM_SCITT_RECEIPTS]
            # For now, assume there is only one receipt
            assert len(parsed_receipts) == 1
            parsed_receipt = parsed_receipts[0]
        else:
            parsed_receipt = cbor2.loads(receipt)

        decoded_receipt = Receipt.from_cose_obj(parsed_receipt)

    service_params = service_trust_store.lookup(decoded_receipt.phdr)
    decoded_receipt.verify(msg, service_params)


class StaticTrustStore(TrustStore):
    """
    A static trust store, based on a list of trusted service certificates.
    """

    services: Dict[str, ServiceParameters]

    def __init__(self, services: Dict[str, ServiceParameters]):
        self.services = services

    @staticmethod
    def load(path: Path) -> "StaticTrustStore":
        """
        Populate a static trust store from a directory. Each JSON file in the
        directory corresponds to a trusted service identity.
        """
        store = {}
        for path in path.glob("**/*.json"):
            with open(path) as f:
                data = json.load(f)

            def _parse_service_param(param: dict):
                service_id = param.get("serviceId")
                if not isinstance(service_id, str) or not service_id:
                    raise ValueError("serviceId must be a non-empty string")

                if service_id in store:
                    raise ValueError(
                        f"Duplicate service ID while reading trust store: {service_id}"
                    )

                store[service_id] = ServiceParameters.from_dict(param)

            # If the JSON file contains a list of parameters, parse each one (as returned by /parameters/historic)
            params = data.get("parameters")
            if params and isinstance(params, list):
                for param in params:
                    _parse_service_param(param)
            # Otherwise, assume the JSON file contains a single set of parameters (as returned by /parameters)
            else:
                _parse_service_param(data)

        return StaticTrustStore(store)

    def lookup(self, phdr) -> ServiceParameters:
        if "service_id" not in phdr:
            raise ValueError("Receipt does not have a service identity.")

        service_id = phdr["service_id"]
        if service_id in self.services:
            return self.services[service_id]
        else:
            raise ValueError(f"Unknown service identity {service_id!r}")


class DIDDocumentTrustStore(TrustStore):
    """
    A trust store backed by a single DID-document.

    The trust store will use the KID found in the protected headers to select
    the appropriate assertion method from the document.
    """

    document: dict

    def __init__(self, document: dict):
        self.document = document

    def lookup(self, phdr) -> ServiceParameters:
        issuer = phdr.get(COSE_HEADER_PARAM_ISSUER)
        if issuer != self.document.get("id"):
            raise ValueError(
                f"Incorrect issuer {issuer!r}, expected {self.document['id']}"
            )

        if pycose.headers.KID in phdr:
            kid = phdr[pycose.headers.KID].decode("ascii")
        else:
            kid = None

        method = did.find_assertion_method(self.document, kid)
        jwk = method["publicKeyJwk"]
        # TODO parse jwk without using x5c as well
        if len(jwk.get("x5c", [])) < 1:
            raise ValueError("Missing x5c parameter in service JWK")
        certificate = base64.b64decode(jwk["x5c"][0])

        return ServiceParameters("CCF", "ES256", certificate)


class DIDResolverTrustStore(TrustStore):
    """
    A trust store which uses the issuer found in receipts to dynamically
    resolve the service parameters.

    The trust store does not restrict which issuers are allowed, only that the
    receipt signature matches the identifier.
    """

    def __init__(self, resolver: Optional[did.Resolver] = None):
        if resolver is not None:
            self.resolver = resolver
        else:
            self.resolver = did.Resolver()

    def lookup(self, phdr) -> ServiceParameters:
        if COSE_HEADER_PARAM_ISSUER not in phdr:
            raise ValueError("Receipt does not have an issuer")

        issuer = phdr[COSE_HEADER_PARAM_ISSUER]
        document = self.resolver.resolve(issuer)

        return DIDDocumentTrustStore(document).lookup(phdr)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from typing import Dict, Optional

from ccf.cose import create_cose_sign1
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKey
from cryptography.hazmat.primitives.serialization import load_pem_private_key

from pyscitt.client import MemberAuthenticationMethod


class LocalKeySignClient(MemberAuthenticationMethod):
    def __init__(self, cert: str, key: str) -> None:
        self.cert = cert
        self.key = key
        self.private_key = load_pem_private_key(
            key.encode("ascii"),
            password=None,
        )

    def cose_sign(self, data: bytes, cose_headers: Optional[Dict] = None) -> bytes:
        """Generates a COSE payload for the specified request with the specified headers.

        https://microsoft.github.io/CCF/main/use_apps/issue_commands.html#signing

        :param data: The intended body for the HTTP request.
        :type data: bytes
        :param cose_headers: The headers to include in the COSE payload.
        :type cose_headers: Optional[Dict]

        :return: The full payload, with signature, to be sent to CCF.
        :rtype: bytes
        """

        assert self.cert, "No identity public certificate available for this identity"
        assert self.key, "No identity private key available for this identity"

        # Use the CCF library to generate the COSE payload
        return create_cose_sign1(
            payload=data,
            key_priv_pem=self.key,
            cert_pem=self.cert,
            additional_protected_header=cose_headers,
        )

    def http_sign(self, data: bytes):
        """
        Generates a HTTP signing payload for the specified request with the specified headers.

        https://microsoft.github.io/CCF/main/use_apps/issue_commands.html#signing

        :param data: The intended body for the HTTP request.
        :type data: bytes

        :return: The full payload, with signature, to be sent to CCF.
        :rtype: bytes
        """

        assert isinstance(self.private_key, (EllipticCurvePrivateKey))
        digest_algo = {256: hashes.SHA256(), 384: hashes.SHA384()}[
            self.private_key.curve.key_size
        ]
        signature = self.private_key.sign(
            signature_algorithm=ec.ECDSA(algorithm=digest_algo), data=data
        )
        return signature



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import argparse
import json
import sys
from dataclasses import dataclass
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional, Union
from urllib.parse import urlparse

import certifi

from pyscitt.did import format_did_web

from .. import governance
from ..client import Client
from .client_arguments import add_client_arguments, create_client


@dataclass
class GovernanceAction:
    proposal: Union[dict, bytes]
    ballot: Optional[str] = None


def add_proposal_arguments(p, *, enable_dump_only=True):
    if enable_dump_only:
        p.add_argument(
            "--dump-only",
            action="store_true",
            help="Dumps the proposal to proposal.json and exits",
        )
    else:
        p.set_defaults(dump_only=False)

    p.add_argument(
        "--allow-open",
        action="store_true",
        help="Do not fail if the proposal remains open",
    )


def execute_action(client: Client, action: GovernanceAction, args: argparse.Namespace):
    if isinstance(action.proposal, bytes):
        proposal = action.proposal
    else:
        proposal = json.dumps(action.proposal).encode("ascii")

    if args.dump_only:
        Path("proposal.json").write_bytes(proposal)
    else:
        result = client.governance.propose(proposal, vote=False)
        if not result.is_accepted:
            result = client.governance.vote(
                result.id, ballot=action.ballot, must_pass=not args.allow_open
            )

        if result.is_accepted:
            print(f"Proposal was accepted!")
        elif result.is_open:
            print(f"Proposal {result.id} is still open, more votes needed!")
        else:
            print(f"Proposal {result.id} is {result.state}")


def propose_root_ca_certs(bundle_name: str, ca_cert_path: Path):
    cert_bundle = ca_cert_path.read_text()
    proposal = governance.set_ca_bundle_proposal(bundle_name, cert_bundle)
    return GovernanceAction(proposal)


def propose_configuration(configuration_path: Path):
    with open(configuration_path) as f:
        configuration = json.load(f)

    proposal = governance.set_scitt_configuration_proposal(configuration)
    return GovernanceAction(proposal)


def propose_open_service(next_service_certificate_path: Path):
    next_service_certificate = next_service_certificate_path.read_text()
    proposal = governance.transition_service_to_open_proposal(next_service_certificate)
    return GovernanceAction(proposal)


def propose_generic(path: Path):
    return GovernanceAction(path.read_bytes())


def propose_constitution(path: Path):
    constitution = path.read_text()
    proposal = governance.set_constitution_proposal(constitution)
    return GovernanceAction(proposal)


SCITT_CONSTITUTION_MARKER_START = "// ----- SCITT Constitution starts here -----\n"
SCITT_CONSTITUTION_MARKER_END = "// ----- SCITT Constitution ends here -----"


def apply_scitt_constitution_update(
    original_constitution: str, new_scitt_constitution: str
) -> str:
    if not new_scitt_constitution.startswith(SCITT_CONSTITUTION_MARKER_START):
        raise RuntimeError(
            f"New SCITT constitution does not start with marker {repr(SCITT_CONSTITUTION_MARKER_START)}"
        )
    if not new_scitt_constitution.rstrip().endswith(SCITT_CONSTITUTION_MARKER_END):
        raise RuntimeError(
            f"New SCITT constitution does not end with marker {repr(SCITT_CONSTITUTION_MARKER_END)}"
        )
    if new_scitt_constitution.count(SCITT_CONSTITUTION_MARKER_START) > 1:
        raise RuntimeError(f"New SCITT constitution contains multiple start markers")
    if new_scitt_constitution.count(SCITT_CONSTITUTION_MARKER_END) > 1:
        raise RuntimeError(f"New SCITT constitution contains multiple end markers")

    parts = original_constitution.split(SCITT_CONSTITUTION_MARKER_START)
    if len(parts) == 1:
        print(
            "Did not find any marker in existing constitution. The SCITT constitution will be appended to it"
        )
    elif len(parts) == 2:
        if not parts[1].rstrip().endswith(SCITT_CONSTITUTION_MARKER_END):
            raise RuntimeError(
                "Existing constitution does not end with the right marker"
            )
    else:
        assert len(parts) > 2
        raise RuntimeError("Found multiple markers in constitution")

    core_constitution = parts[0]
    if not core_constitution.endswith("\n"):
        core_constitution += "\n"

    return core_constitution + new_scitt_constitution


def update_scitt_constitution(client: Client, scitt_constitution_path: Path, yes: bool):
    scitt_constitution = scitt_constitution_path.read_text()
    initial_constitution = client.get_constitution()
    final_constitution = apply_scitt_constitution_update(
        initial_constitution, scitt_constitution
    )

    if not yes:
        with TemporaryDirectory() as d:
            path = Path(d).joinpath("constitution.js")
            path.write_text(final_constitution)

            print(f"Updated constitution was written to {path}.")
            print("You may review and modify the file before proceeding.")
            answer = input("Do you wish to continue [y/N]: ").lower()

            if answer == "y" or answer == "yes":
                final_constitution = path.read_text()
                pass
            else:
                print("Aborting")
                sys.exit(1)

    proposal = governance.set_constitution_proposal(final_constitution)

    # The ballot we submit will only approve the proposal if the constitution
    # found in the KV still matches the original value we got before applying the
    # update. Otherwise a concurrent change has happened, and we shouldn't be making
    # any changes.
    # Note the use of `repr` to turn the constitution into a valid Javascript
    # string literal, in particular, escaping quotations marks and newlines.
    ballot = f"""
        export function vote (rawProposal, proposerId) {{
            const singletonKey = new ArrayBuffer(8);
            const constitution = ccf.bufToJsonCompatible(ccf.kv["public:ccf.gov.constitution"].get(singletonKey));
            return constitution == {repr(initial_constitution)};
        }}
    """
    return GovernanceAction(proposal, ballot)


def get_constitution(client: Client, path: Path):
    path.write_text(client.get_constitution())


def setup_local_development(
    client: Client, trust_store_dir: Optional[Path], did_web_ca_certs: Optional[Path]
):
    # Member activation and opening the service are not necessary when
    # using sandbox.sh, since that will already been taken care of.
    # They are idempotent operation anyway, so no harm in doing them again.

    print("Activating member credentials...")
    client.governance.activate_member()

    print("Configuring DID Web root CA bundle...")
    if did_web_ca_certs:
        bundle = did_web_ca_certs.read_text()
    else:
        # This uses the Mozilla root program. certifi is the same package that
        # provides roots for eg. the requests and httpx modules.
        bundle = certifi.contents()
    proposal = governance.set_ca_bundle_proposal("did_web_tls_roots", bundle)
    client.governance.propose(proposal, must_pass=True)

    print("Configuring service...")
    url = urlparse(client.url)
    assert url.hostname is not None

    config = {
        "authentication": {"allow_unauthenticated": True},
        "service_identifier": format_did_web(url.hostname, url.port),
    }
    proposal = governance.set_scitt_configuration_proposal(config)
    client.governance.propose(proposal, must_pass=True)

    print("Opening service...")
    network = client.get("node/network").json()
    proposal = governance.transition_service_to_open_proposal(
        network["service_certificate"]
    )
    client.governance.propose(proposal, must_pass=True)
    client.wait_for_network_open()

    if trust_store_dir:
        print(f"Adding service to {trust_store_dir} ...")
        trust_store_dir.mkdir(parents=True, exist_ok=True)
        parameters = client.get_parameters().as_dict()
        service_id = parameters["serviceId"]
        path = trust_store_dir.joinpath(service_id + ".json")
        path.write_text(json.dumps(parameters))


def cli(fn):
    parser = fn(description="Execute governance actions")
    sub = parser.add_subparsers(help="Governance action to execute", required=True)

    p = sub.add_parser("propose_ca_certs")
    add_client_arguments(p, with_member_auth=True)
    add_proposal_arguments(p)
    p.add_argument(
        "--name",
        choices=["did_web_tls_roots", "x509_roots"],
        required=True,
        help="Name of the CA Cert bundle as referenced in CCF apps",
        dest="bundle_name",
    )
    p.add_argument(
        "--ca-certs",
        type=Path,
        help="Path to PEM-encoded CA Cert bundle",
        required=True,
    )
    p.set_defaults(
        func=lambda args: execute_action(
            create_client(args),
            propose_root_ca_certs(args.bundle_name, args.ca_certs),
            args,
        )
    )

    p = sub.add_parser("propose_configuration")
    add_client_arguments(p, with_member_auth=True)
    add_proposal_arguments(p)
    p.add_argument(
        "--configuration", type=Path, help="JSON configuration", required=True
    )
    p.set_defaults(
        func=lambda args: execute_action(
            create_client(args),
            propose_configuration(args.configuration),
            args,
        )
    )

    p = sub.add_parser("propose_generic")
    add_client_arguments(p, with_member_auth=True)
    add_proposal_arguments(p, enable_dump_only=False)
    p.add_argument("--proposal-path", type=Path, help="JSON proposal", required=True)
    p.set_defaults(
        func=lambda args: execute_action(
            create_client(args),
            propose_generic(args.proposal_path),
            args,
        )
    )

    p = sub.add_parser("propose_open_service")
    add_client_arguments(p, with_member_auth=True)
    add_proposal_arguments(p)
    p.add_argument(
        "--next-service-certificate",
        type=Path,
        help="Path to next service certificate",
        required=True,
    )
    p.set_defaults(
        func=lambda args: execute_action(
            create_client(args),
            propose_open_service(args.next_service_certificate),
            args,
        )
    )

    p = sub.add_parser(
        "propose_constitution", help="Propose a new service constitution"
    )
    add_client_arguments(p, with_member_auth=True)
    add_proposal_arguments(p)
    p.add_argument(
        "--constitution-file",
        type=Path,
        required=True,
        help="Path to the new constitution",
    )
    p.set_defaults(
        func=lambda args: execute_action(
            create_client(args),
            propose_constitution(args.constitution_file),
            args,
        )
    )

    p = sub.add_parser(
        "update_scitt_constitution",
        help="Update the SCITT component of the service's constitution",
    )
    add_client_arguments(p, with_member_auth=True)
    add_proposal_arguments(p)
    p.add_argument(
        "--scitt-constitution-file",
        type=Path,
        required=True,
        help="Path to the scitt.js file which contains the new SCITT component of the constitution",
    )
    p.add_argument("--yes", action="store_true", help="Do not ask for confirmation")

    def func(args):
        client = create_client(args)
        action = update_scitt_constitution(
            client, args.scitt_constitution_file, args.yes
        )
        execute_action(client, action, args)

    p.set_defaults(func=func)

    p = sub.add_parser("activate_member")
    add_client_arguments(p, with_member_auth=True)
    p.set_defaults(func=lambda args: create_client(args).governance.activate_member())

    p = sub.add_parser(
        "constitution",
        help="Fetch the service's current constitution and write it to a file",
    )
    add_client_arguments(p)
    p.add_argument("--output", type=Path, help="Output file", required=True)
    p.set_defaults(func=lambda args: get_constitution(create_client(args), args.output))

    p = sub.add_parser(
        "local_development", help="Configure a SCITT ledger for local development."
    )
    add_client_arguments(p, with_member_auth=True, development_only=True)
    p.add_argument(
        "--service-trust-store",
        type=Path,
        help="Folder containing a SCITT trust store, to which this service's identity will be added.",
    )
    p.add_argument(
        "--did-web-ca-certs",
        type=Path,
        help="Path to PEM-encoded CA Cert bundle, used by the service when resolving DID web issuers. By default, the Mozilla CA roots are used.",
    )
    p.set_defaults(
        func=lambda args: setup_local_development(
            create_client(args), args.service_trust_store, args.did_web_ca_certs
        )
    )

    return parser


if __name__ == "__main__":
    parser = cli(argparse.ArgumentParser)
    args = parser.parse_args()
    args.func(args)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from os import path

from setuptools import find_packages, setup

PACKAGE_NAME = "pyscitt"
PACKAGE_VERSION = "0.3.1"

path_here = path.abspath(path.dirname(__file__))

with open(path.join(path_here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name=PACKAGE_NAME,
    version=PACKAGE_VERSION,
    description="Tools to sign claims and interact with a SCITT CCF Ledger",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    entry_points={
        "console_scripts": ["scitt=pyscitt.cli.main:main"],
    },
    python_requires=">=3.8",
    install_requires=[
        "ccf==5.0.0",
        "cryptography==42.*",  # needs to match ccf
        "httpx",
        "cbor2==5.4.*",
        "pycose==1.1.0",
        "pyjwt",
        "azure-keyvault",
        "azure-identity",
    ],
    license="Apache License 2.0",
    author="SCITT CCF Team",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    url="https://github.com/microsoft/scitt-ccf-ledger/tree/main/pyscitt",
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from typing import Optional, Tuple
from urllib.parse import quote, unquote

import httpx
from loguru import logger as LOG

from .crypto import Pem, Signer, get_public_key_fingerprint, jwk_from_public_key

DID_FILENAME = "did.json"
DID_WEB_DOC_URL_SCHEME = "https"
DID_WEB_DOC_WELLKNOWN_PATH = ".well-known"
DID_WEB_PREFIX = "did:web:"


def format_did_web(
    host: str,
    port: Optional[int],
    path: Optional[str] = None,
) -> str:
    did = DID_WEB_PREFIX + host
    if port:
        did += quote(f":{port}")
    if path:
        assert not path.startswith("/")
        did += ":" + path.replace("/", ":")

    return did


def did_web_parse(did: str) -> Tuple[str, Optional[str]]:
    parts = did.split(":")

    if len(parts) < 3:
        raise ValueError("Malformed DID-web")

    prefix, method, location = parts[:3]
    if prefix != "did" or method != "web":
        raise ValueError("Malformed DID-web")

    if len(parts) == 3:
        path = None
    else:
        path = "/".join(parts[3:])

    return unquote(location), path


def did_web_document_url(did: str) -> str:
    (location, path) = did_web_parse(did)
    if path is None:
        path = DID_WEB_DOC_WELLKNOWN_PATH

    return f"{DID_WEB_DOC_URL_SCHEME}://{location}/{path}/{DID_FILENAME}"


def find_assertion_method(did_doc: dict, kid: Optional[str]):
    # TODO: support non-inline verification methods
    assertion_methods = did_doc["assertionMethod"]
    if kid is None:
        if len(assertion_methods) > 1:
            raise ValueError("found more than one assertion method")
        elif len(assertion_methods) == 0:
            raise ValueError("no assertion method found")
        else:
            return assertion_methods[0]

    else:
        matches = [
            m for m in assertion_methods if get_verification_method_kid(m) == kid
        ]
        if len(matches) > 1:
            raise ValueError("found more than one assertion method with given kid")
        elif len(matches) == 0:
            raise ValueError("no assertion method found with given kid")
        return matches[0]


# The "kid" COSE header parameter refers
# to the fragment of the verification method's ID.
# The part before the fragment is the DID itself
# and stored in "issuer".
def get_verification_method_kid(obj: dict) -> str:
    return "#" + obj["id"].split("#")[1]


def get_signer(private_key: Pem, did_doc: dict, kid: Optional[str] = None) -> Signer:
    assertion_method = find_assertion_method(did_doc, kid)
    kid = get_verification_method_kid(assertion_method)
    algorithm = assertion_method["publicKeyJwk"]["alg"]

    return Signer(private_key, did_doc["id"], kid, algorithm)


def create_assertion_method(
    *,
    did: str,
    public_key: Pem,
    alg: Optional[str] = None,
    kid: Optional[str] = None,
):
    if kid is None:
        kid = "#" + get_public_key_fingerprint(public_key)
    if not kid.startswith("#"):
        raise ValueError("kid must start with '#'")

    return {
        "id": f"{did}{kid}",
        "type": "JsonWebKey2020",
        "controller": did,
        "publicKeyJwk": jwk_from_public_key(public_key, alg, kid[1:]),
    }


def create_document(
    *,
    did: str,
    assertion_methods: Optional[list] = None,
    public_key: Optional[Pem] = None,
) -> dict:
    if not did.startswith("did:"):
        raise ValueError("did must start with 'did:'")

    if assertion_methods is None:
        if public_key is None:
            raise ValueError(
                "Either a public key or assertion methods must be provided"
            )
        assertion_methods = [create_assertion_method(did=did, public_key=public_key)]

    return {
        "@context": [
            "https://www.w3.org/ns/did/v1",
            "https://w3id.org/security/suites/jws-2020/v1",
        ],
        "id": did,
        "assertionMethod": assertion_methods,
    }


class Resolver:
    def __init__(self, *, verify: bool = True):
        self.verify = verify

    def resolve(self, did: str) -> dict:
        parts = did.split(":", 3)
        if len(parts) < 3:
            raise ValueError("Malformed DID")

        if parts[0] != "did":
            raise ValueError("Malformed DID")
        if parts[1] != "web":
            raise ValueError(f"Unsupported DID method {parts[1]!r}")

        url = did_web_document_url(did)

        LOG.debug(f"Resolving {did!r} at {url!r}")
        r = httpx.get(url, verify=self.verify)
        r.raise_for_status()
        return r.json()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import argparse
from pathlib import Path
from typing import Optional

from ..client import Client
from ..verify import StaticTrustStore, verify_receipt
from .client_arguments import add_client_arguments, create_client


def retrieve_signed_claimsets(
    client: Client,
    base_path: Path,
    from_seqno: Optional[int],
    to_seqno: Optional[int],
    service_trust_store_path: Optional[Path],
    embed_receipt: bool = False,
):
    base_path.mkdir(parents=True, exist_ok=True)

    if service_trust_store_path:
        service_trust_store = StaticTrustStore.load(service_trust_store_path)
    else:
        service_trust_store = None

    for tx in client.enumerate_claims(start=from_seqno, end=to_seqno):
        claim = client.get_claim(tx, embed_receipt=embed_receipt)
        path = base_path / f"{tx}.cose"

        if service_trust_store:
            verify_receipt(claim, service_trust_store=service_trust_store)

        with open(path, "wb") as f:
            f.write(claim)


def cli(fn):
    parser = fn(
        description="Retrieve signed claimsets from a SCITT CCF Ledger together with receipts"
    )
    add_client_arguments(parser)
    parser.add_argument(
        "path", type=Path, help="Folder to store signed claimsets and receipts"
    )
    parser.add_argument(
        "--from", dest="from_seqno", type=int, help="Start seqno (optional)"
    )
    parser.add_argument("--to", dest="to_seqno", type=int, help="End seqno (optional)")
    parser.add_argument(
        "--service-trust-store",
        type=Path,
        help="Folder containing JSON parameter files of SCITT services to trust",
    )
    parser.add_argument(
        "--embed-receipt",
        action="store_true",
        default=False,
        help="Whether to include the receipt in the signed claimset under the unprotected header. Default is False.",
    )

    def cmd(args):
        client = create_client(args)
        retrieve_signed_claimsets(
            client,
            args.path,
            args.from_seqno,
            args.to_seqno,
            args.service_trust_store,
            args.embed_receipt,
        )

    parser.set_defaults(func=cmd)
    return parser


if __name__ == "__main__":
    parser = cli(argparse.ArgumentParser)
    args = parser.parse_args()
    args.func(args)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import argparse
import json
from pathlib import Path
from typing import Optional
from urllib.parse import urlsplit

from .. import crypto, did
from ..did import did_web_document_url, format_did_web

DID_FILENAME = "did.json"


def write_file(path: Path, contents: str):
    print(f"Writing {path}")
    path.write_text(contents)


def create_did_web(
    base_url: str,
    out_dir: Path,
    ssh_key_path: Optional[Path],
    kty: str,
    alg: Optional[str],
):
    parsed = urlsplit(base_url)
    assert parsed.hostname
    identifier = format_did_web(
        host=parsed.hostname, port=parsed.port, path=parsed.path.lstrip("/")
    )

    did_doc_url = did_web_document_url(identifier)
    print(f"did:web document URL: {did_doc_url}")

    out_path = out_dir / DID_FILENAME
    if out_path.exists():
        return

    out_dir.mkdir(parents=True, exist_ok=True)

    if ssh_key_path:
        ssh_key = ssh_key_path.read_text()
        public_key_pem = crypto.ssh_public_key_to_pem(ssh_key)
    else:
        public_key_pem_path = out_dir / "key_pub.pem"
        private_key_pem_path = out_dir / "key.pem"
        if not public_key_pem_path.exists():
            private_key_pem, public_key_pem = crypto.generate_keypair(kty)
            write_file(private_key_pem_path, private_key_pem)
            write_file(public_key_pem_path, public_key_pem)
        else:
            public_key_pem = public_key_pem_path.read_text()

    method = did.create_assertion_method(
        did=identifier, public_key=public_key_pem, alg=alg
    )
    doc = did.create_document(did=identifier, assertion_methods=[method])
    write_file(out_path, json.dumps(doc, indent=2))


def cli(fn):
    parser = fn(description="Create a DID document and keys for a did:web identity")
    parser.add_argument("--url", required=True, help="Base URL")
    parser.add_argument(
        "--ssh-key",
        type=Path,
        help="Path to existing SSH public key instead of generating a new key pair",
    )
    parser.add_argument("--kty", choices=["ec", "rsa"], default="ec", help="Key type")
    parser.add_argument(
        "--alg", help="Signing algorithm the key is intended to be used with."
    )
    parser.add_argument(
        "--out-dir", required=True, type=Path, help="Path to store output files"
    )

    def cmd(args):
        create_did_web(args.url, args.out_dir, args.ssh_key, args.kty, args.alg)

    parser.set_defaults(func=cmd)

    return parser


if __name__ == "__main__":
    parser = cli(argparse.ArgumentParser)
    args = parser.parse_args()
    args.func(args)
